import logging
import os
from mock import patch

import pytest

from sciencebeam_utils.tools import find_file_pairs
from sciencebeam_utils.tools.find_file_pairs import (
    to_relative_file_pairs,
    run,
    parse_args,
    main
)


LOGGER = logging.getLogger(__name__)

BASE_SOURCE_PATH = '/source'

PDF_FILE_1 = BASE_SOURCE_PATH + '/file1.pdf'
XML_FILE_1 = BASE_SOURCE_PATH + '/file1.xml'
PDF_FILE_2 = BASE_SOURCE_PATH + '/file2.pdf'
XML_FILE_2 = BASE_SOURCE_PATH + '/file2.xml'

SOURCE_PATTERN = '*.pdf'
XML_PATTERN = '*.xml'
OUTPUT_FILE = 'file-list.tsv'

SOME_ARGV = [
    '--data-path=%s' % BASE_SOURCE_PATH,
    '--source-pattern=%s' % SOURCE_PATTERN,
    '--xml-pattern=%s' % XML_PATTERN,
    '--out=%s' % OUTPUT_FILE
]


@pytest.fixture(name='to_relative_file_pairs_mock')
def _to_relative_file_pairs():
    with patch.object(find_file_pairs, 'to_relative_file_pairs') as m:
        yield m


@pytest.fixture(name='find_file_pairs_grouped_by_parent_directory_or_name_mock')
def _find_file_pairs_grouped_by_parent_directory_or_name():
    with patch.object(find_file_pairs, 'find_file_pairs_grouped_by_parent_directory_or_name') as m:
        m.return_value = [
            (PDF_FILE_1, XML_FILE_1),
            (PDF_FILE_2, XML_FILE_2)
        ]
        yield m


@pytest.fixture(name='save_file_pairs_to_csv_mock')
def _save_file_pairs_to_csv():
    with patch.object(find_file_pairs, 'save_file_pairs_to_csv') as m:
        yield m


@pytest.fixture(name='parse_args_mock')
def _parse_args():
    with patch.object(find_file_pairs, 'parse_args') as m:
        yield m


@pytest.fixture(name='run_mock')
def _run():
    with patch.object(find_file_pairs, 'run') as m:
        yield m


def _touch(path):
    path.write(b'', ensure=True)
    return path


@pytest.fixture(name='pdf_file_1')
def _pdf_file_1(tmpdir):
    return _touch(tmpdir.join(PDF_FILE_1))


@pytest.fixture(name='xml_file_1')
def _xml_file_1(tmpdir):
    return _touch(tmpdir.join(XML_FILE_1))


@pytest.fixture(name='data_path')
def _data_path(tmpdir):
    return tmpdir.join(BASE_SOURCE_PATH)


@pytest.fixture(name='out_file')
def _out_file(tmpdir):
    return tmpdir.join(OUTPUT_FILE)


class TestToRelativeFilePairs:
    def test_should_make_paths_relative(self):
        assert list(to_relative_file_pairs(
            '/parent',
            [('/parent/sub/file1', '/parent/sub/file2')]
        )) == [('sub/file1', 'sub/file2')]


class TestRun:
    def test_should_pass_around_parameters(
            self,
            find_file_pairs_grouped_by_parent_directory_or_name_mock,
            save_file_pairs_to_csv_mock):

        opt = parse_args(SOME_ARGV)
        run(opt)
        find_file_pairs_grouped_by_parent_directory_or_name_mock.assert_called_with([
            os.path.join(BASE_SOURCE_PATH, SOURCE_PATTERN),
            os.path.join(BASE_SOURCE_PATH, XML_PATTERN)
        ])
        save_file_pairs_to_csv_mock.assert_called_with(
            opt.out,
            find_file_pairs_grouped_by_parent_directory_or_name_mock.return_value
        )

    def test_should_use_relative_paths_if_enabled(
            self,
            find_file_pairs_grouped_by_parent_directory_or_name_mock,
            to_relative_file_pairs_mock,
            save_file_pairs_to_csv_mock):

        opt = parse_args(SOME_ARGV)
        opt.use_relative_paths = True

        to_relative_file_pairs_mock.return_value = [('file1.pdf', 'file1.xml')]

        run(opt)
        to_relative_file_pairs_mock.assert_called_with(
            BASE_SOURCE_PATH,
            find_file_pairs_grouped_by_parent_directory_or_name_mock.return_value
        )
        save_file_pairs_to_csv_mock.assert_called_with(
            opt.out,
            to_relative_file_pairs_mock.return_value
        )

    def test_should_generate_file_list(self, data_path, pdf_file_1, xml_file_1, out_file):
        LOGGER.debug('pdf_file_1: %s, xml_file: %s', pdf_file_1, xml_file_1)
        opt = parse_args(SOME_ARGV)
        opt.data_path = str(data_path)
        opt.out = str(out_file)
        run(opt)
        out_lines = [s.strip() for s in out_file.read().strip().split('\n')]
        LOGGER.debug('out_lines: %s', out_lines)
        assert out_lines == [
            'source_url\txml_url',
            '%s\t%s' % (pdf_file_1, xml_file_1)
        ]


class TestMain:
    def test_should_parse_args_and_call_run(self, parse_args_mock, run_mock):
        main(SOME_ARGV)
        parse_args_mock.assert_called_with(SOME_ARGV)
        run_mock.assert_called_with(parse_args_mock.return_value)



import pickle


class SerializableException(RuntimeError):
    pass


def is_serializable(obj) -> bool:
    try:
        pickle.dumps(obj)
        return True
    except Exception:  # pylint: disable=broad-except
        return False


def get_serializable_exception(exception: BaseException) -> BaseException:
    if is_serializable(exception):
        return exception
    serializable_exception = SerializableException(str(exception))
    serializable_exception.__cause__ = get_serializable_exception(exception.__cause__)
    serializable_exception.__context__ = get_serializable_exception(exception.__context__)
    serializable_exception.__traceback__ = exception.__traceback__
    return serializable_exception



from __future__ import division

import argparse
import logging
from concurrent.futures import ThreadPoolExecutor

from apache_beam.io.filesystems import FileSystems

from sciencebeam_utils.utils.file_list import (
    load_file_list
)

from sciencebeam_utils.tools.tool_utils import (
  setup_logging,
  add_limit_args,
  add_default_args,
  process_default_args
)


LOGGER = logging.getLogger(__name__)


def parse_args(argv=None):
    parser = argparse.ArgumentParser(
        'Check file list'
    )

    source = parser.add_argument_group('source')
    source.add_argument(
        '--file-list', type=str, required=True,
        help='path to source file list (tsv/csv/lst)'
    )
    source.add_argument(
        '--file-column', type=str, required=False,
        default='url',
        help='csv/tsv column (ignored for plain file list)'
    )

    add_limit_args(parser)
    add_default_args(parser)

    return parser.parse_args(argv)


def map_file_list_to_file_exists(file_list):
    with ThreadPoolExecutor(max_workers=50) as executor:
        return list(executor.map(FileSystems.exists, file_list))


def format_file_list(file_list):
    return '%s' % file_list


def format_file_exists_results(file_exists, file_list):
    if not file_exists:
        return 'empty file list'
    file_exists_count = sum(file_exists)
    file_missing_count = len(file_exists) - file_exists_count
    files_missing = [s for s, exists in zip(file_list, file_exists) if not exists]
    return (
        'files exist: %d (%.0f%%), files missing: %d (%.0f%%)%s' % (
            file_exists_count, 100.0 * file_exists_count / len(file_exists),
            file_missing_count, 100.0 * file_missing_count / len(file_exists),
            (
                ' (example missing: %s)' % format_file_list(files_missing[:3])
                if files_missing
                else ''
            )
        )
    )


def check_files_and_report_result(file_list):
    file_exists = map_file_list_to_file_exists(file_list)
    LOGGER.info('%s', format_file_exists_results(file_exists, file_list))
    assert sum(file_exists) > 0


def run(opt):
    file_list = load_file_list(
        opt.file_list,
        column=opt.file_column,
        limit=opt.limit
    )
    check_files_and_report_result(file_list)


def main(argv=None):
    args = parse_args(argv)

    process_default_args(args)

    run(args)


if __name__ == '__main__':
    setup_logging()

    main()



import logging
import os
from functools import reduce  # pylint: disable=redefined-builtin
from typing import Iterable, List, Tuple

from apache_beam.io.filesystems import FileSystems

from sciencebeam_utils.utils.collection import (
    groupby_to_dict,
    sort_and_groupby_to_dict
)

from .file_path import strip_ext


LOGGER = logging.getLogger(__name__)


def find_matching_filenames(pattern):
    return (x.path for x in FileSystems.match([pattern])[0].metadata_list)


def group_files_by_parent_directory(filenames):
    return groupby_to_dict(sorted(filenames), os.path.dirname)


def group_files_by_name_excl_ext(filenames):
    return sort_and_groupby_to_dict(filenames, strip_ext)


def zip_by_keys(*dict_list):
    keys = reduce(lambda agg, v: agg | set(v.keys()), dict_list, set())
    return (
        [d.get(k) for d in dict_list]
        for k in sorted(keys)
    )


def group_file_pairs_by_parent_directory_or_name(
    files_by_type: List[List[str]]
) -> Iterable[Tuple[str, ...]]:
    grouped_files_by_pattern = [
        group_files_by_parent_directory(files) for files in files_by_type
    ]
    for files_in_group_by_pattern in zip_by_keys(*grouped_files_by_pattern):
        if all(len(files or []) == 1 for files in files_in_group_by_pattern):
            yield tuple([  # pylint: disable=consider-using-generator
                files[0] for files in files_in_group_by_pattern
            ])
        else:
            grouped_by_name = [
                group_files_by_name_excl_ext(files or [])
                for files in files_in_group_by_pattern
            ]
            for files_by_name in zip_by_keys(*grouped_by_name):
                if all(len(files or []) == 1 for files in files_by_name):
                    yield tuple([  # pylint: disable=consider-using-generator
                        files[0] for files in files_by_name
                    ])
                else:
                    LOGGER.info(
                        'no exclusively matching files found: %s',
                        list(files_by_name)
                    )


def find_file_pairs_grouped_by_parent_directory_or_name(patterns):
    matching_files_by_pattern = [
        list(find_matching_filenames(pattern)) for pattern in patterns
    ]
    LOGGER.info(
        'found number of files %s',
        ', '.join(
            '%s: %d' % (pattern, len(files))
            for pattern, files in zip(patterns, matching_files_by_pattern)
        )
    )
    patterns_without_files = [
        pattern
        for pattern, files in zip(patterns, matching_files_by_pattern)
        if len(files) == 0
    ]
    if patterns_without_files:
        raise RuntimeError('no files found for: %s' % patterns_without_files)
    return group_file_pairs_by_parent_directory_or_name(
        matching_files_by_pattern
    )



import pytest

from sciencebeam_utils.utils.file_path import (
    relative_path,
    join_if_relative_path,
    change_ext,
    get_output_file,
    base_path_for_file_list,
    get_or_validate_base_path
)


class TestRelativePath:
    def test_should_return_path_if_base_path_is_none(self):
        assert relative_path(None, 'file') == 'file'

    def test_should_return_path_if_path_outside_base_path(self):
        assert relative_path('/parent', '/other/file') == '/other/file'

    def test_should_return_absolute_path_if_base_path_matches(self):
        assert relative_path('/parent', '/parent/file') == 'file'


class TestJoinIfRelativePath:
    def test_should_return_path_if_base_path_is_none(self):
        assert join_if_relative_path(None, 'file') == 'file'

    def test_should_return_path_if_not_relative(self):
        assert join_if_relative_path('/parent', '/other/file') == '/other/file'

    def test_should_return_joined_path_if_relative(self):
        assert join_if_relative_path('/parent', 'file') == '/parent/file'


class TestChangeExt:
    def test_should_replace_simple_ext_with_simple_ext(self):
        assert change_ext('file.pdf', None, '.xml') == 'file.xml'

    def test_should_replace_simple_ext_with_combined_ext(self):
        assert change_ext('file.pdf', None, '.svg.zip') == 'file.svg.zip'

    def test_should_remove_gz_ext_before_replacing_ext(self):
        assert change_ext('file.pdf.gz', None, '.svg.zip') == 'file.svg.zip'


class TestGetOutputFile:
    def test_should_return_output_file_with_path_and_change_ext(self):
        assert get_output_file(
            '/source/path/file.pdf',
            '/source',
            '/output',
            '.xml'
        ) == '/output/path/file.xml'


class TestBasePathForFileList:
    def test_should_return_empty_string_if_file_list_is_empty(self):
        assert base_path_for_file_list([]) == ''

    def test_should_return_empty_string_if_filename_is_empty(self):
        assert base_path_for_file_list(['']) == ''

    def test_should_return_parent_directory_of_single_file(self):
        assert base_path_for_file_list(['/base/path/1/file']) == '/base/path/1'

    def test_should_return_common_path_of_two_files(self):
        assert base_path_for_file_list(['/base/path/1/file', '/base/path/2/file']) == '/base/path'

    def test_should_return_common_path_of_two_files_using_protocol(self):
        assert base_path_for_file_list([
            'a://base/path/1/file', 'a://base/path/2/file'
        ]) == 'a://base/path'

    def test_should_return_common_path_of_two_files_using_forward_slash(self):
        assert base_path_for_file_list([
            '\\base\\path\\1\\file', '\\base\\path\\2\\file'
        ]) == '\\base\\path'

    def test_should_return_empty_string_if_no_common_path_was_found(self):
        assert base_path_for_file_list(['a://base/path/1/file', 'b://base/path/2/file']) == ''

    def test_should_return_common_path_ignoring_partial_name_match(self):
        assert base_path_for_file_list(['/base/path/file1', '/base/path/file2']) == '/base/path'


class TestGetOrValidateBasePath:
    def test_should_return_base_path_of_two_files_if_no_base_path_was_provided(self):
        assert get_or_validate_base_path(
            ['/base/path/1/file', '/base/path/2/file'],
            None
        ) == '/base/path'

    def test_should_return_passed_in_base_path_if_valid(self):
        assert get_or_validate_base_path(
            ['/base/path/1/file', '/base/path/2/file'],
            '/base'
        ) == '/base'

    def test_should_raise_error_if_passed_in_base_path_is_invalid(self):
        with pytest.raises(AssertionError):
            get_or_validate_base_path(
                ['/base/path/1/file', '/base/path/2/file'],
                '/base/other'
            )



import logging

from tqdm import tqdm


LOGGER = logging.getLogger(__name__)


class logging_tqdm(tqdm):
    def __init__(
            self,
            *args,
            logger: logging.Logger = None,
            mininterval: float = 1,
            bar_format: str = '{desc}{percentage:3.0f}%{r_bar}',
            desc: str = 'progress: ',
            **kwargs):
        self._logger = logger
        super().__init__(
            *args,
            mininterval=mininterval,
            bar_format=bar_format,
            desc=desc,
            **kwargs
        )

    @property
    def logger(self):
        if self._logger is not None:
            return self._logger
        return LOGGER

    def display(self, msg=None, pos=None):
        if not self.n:
            # skip progress bar before having processed anything
            return
        if not msg:
            msg = self.__str__()
        self.logger.info('%s', msg)



import errno
import logging
import os
import subprocess
from getpass import getuser
from time import gmtime, strftime


from six import text_type


def get_logger():
    return logging.getLogger(__name__)


def create_fn_api_runner():
    # pylint: disable=import-outside-toplevel
    from apache_beam.runners.portability.fn_api_runner import FnApiRunner
    return FnApiRunner()


def get_cloud_project():
    cmd = [
        'gcloud', '-q', 'config', 'list', 'project',
        '--format=value(core.project)'
    ]
    with open(os.devnull, 'w') as dev_null:
        try:
            res = subprocess.check_output(cmd, stderr=dev_null).strip()
            if not res:
                raise Exception(
                    '--cloud specified but no Google Cloud Platform '
                    'project found.\n'
                    'Please specify your project name with the --project '
                    'flag or set a default project: '
                    'gcloud config set project YOUR_PROJECT_NAME'
                )

            if not isinstance(res, text_type):
                res = res.decode('utf-8')
            return res
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise Exception(
                    'gcloud is not installed. The Google Cloud SDK is '
                    'necessary to communicate with the Cloud ML service. '
                    'Please install and set up gcloud.'
                ) from e
            raise


def get_default_job_name(name, suffix=''):
    timestamp_str = strftime("%Y%m%d-%H%M%S", gmtime())
    return '%s-%s%s-%s' % (name or 'beamapp', getuser(), suffix or '', timestamp_str)


def add_cloud_args(parser):
    parser.add_argument(
        '--cloud',
        default=False,
        action='store_true'
    )
    parser.add_argument(
        '--runner',
        required=False,
        default=None,
        help='Runner.'
    )
    parser.add_argument(
        '--project',
        type=str,
        help='The cloud project name to be used for running this pipeline'
    )
    parser.add_argument(
        '--num-workers', '--num_workers',
        default=1,
        type=int,
        help='The number of workers.'
    )
    parser.add_argument(
        '--max-workers', '--max_num_workers',
        dest='max_num_workers',
        type=int,
        help='The number of maximum workers (with auto-scaling).'
    )
    parser.add_argument(
        '--job-name', '--job_name',
        type=str, required=False,
        help='The name of the cloud job'
    )
    parser.add_argument(
        '--job-name-suffix', type=str, required=False,
        help='A suffix appended to the job name'
    )
    return parser


def process_cloud_args(parsed_args, output_path, name=None):
    if parsed_args.num_workers:
        parsed_args.autoscaling_algorithm = 'NONE'
        parsed_args.max_num_workers = parsed_args.num_workers
    parsed_args.setup_file = './setup.py'

    if parsed_args.cloud:
        # Flags which need to be set for cloud runs.
        default_values = {
            'temp_location':  os.path.join(os.path.dirname(output_path), 'temp'),
            'runner': 'DataflowRunner',
            'save_main_session': True,
        }
        if not parsed_args.project:
            parsed_args.project = get_cloud_project()
        if not parsed_args.job_name:
            parsed_args.job_name = get_default_job_name(name, parsed_args.job_name_suffix)
    else:
        # Flags which need to be set for local runs.
        default_values = {
            'runner': 'DirectRunner',
        }

    get_logger().info('default_values: %s', default_values)
    for kk, vv in default_values.items():
        if kk not in parsed_args or not vars(parsed_args)[kk]:
            vars(parsed_args)[kk] = vv

    if parsed_args.runner == 'FnApiRunner':
        parsed_args.runner = create_fn_api_runner()



import logging

import pytest

import apache_beam as beam
from apache_beam.testing.util import (
    assert_that,
    equal_to
)

from sciencebeam_utils.beam_utils.testing import (
    BeamTest,
    TestPipeline,
    get_counter_value
)

from sciencebeam_utils.beam_utils.utils import (
    MapOrLog,
    TransformAndLog,
    TransformAndCount,
    PreventFusion
)


SOME_VALUE_1 = 'value 1'
SOME_VALUE_2 = 'value 2'
SOME_VALUE_CAUSING_EXCEPTION = 1


def SOME_FN(x):
    return x.upper()


def FN_RAISING_EXCEPTION(_):
    raise RuntimeError('oh dear')


ERROR_COUNT_METRIC_NAME = 'error_count'
COUNT_METRIC_NAME_1 = 'count_1'


def get_logger():
    return logging.getLogger(__name__)


def setup_module():
    logging.basicConfig(level='DEBUG')


@pytest.mark.slow
class TestMapOrLog(BeamTest):
    def test_should_pass_through_return_value_if_no_exception_was_raised(self):
        def fn(x):
            return x.upper()
        with TestPipeline() as p:
            result = (
                p |
                beam.Create([SOME_VALUE_1]) |
                MapOrLog(SOME_FN)
            )
            assert_that(result, equal_to([fn(SOME_VALUE_1)]))

    def test_should_skip_entries_that_cause_an_exception(self):
        with TestPipeline() as p:
            result = (
                p |
                beam.Create([SOME_VALUE_1]) |
                MapOrLog(FN_RAISING_EXCEPTION)
            )
            assert_that(result, equal_to([]))

    def test_should_not_increase_error_metric_counter_if_no_exception_raised(self):
        with TestPipeline() as p:
            _ = (  # noqa: F841
                p |
                beam.Create([SOME_VALUE_1]) |
                MapOrLog(FN_RAISING_EXCEPTION, error_count=ERROR_COUNT_METRIC_NAME)
            )
            assert get_counter_value(p.run(), ERROR_COUNT_METRIC_NAME) == 1

    def test_should_increase_error_metric_counter_if_exception_was_raised(self):
        with TestPipeline() as p:
            _ = (  # noqa: F841
                p |
                beam.Create([SOME_VALUE_1]) |
                MapOrLog(FN_RAISING_EXCEPTION, error_count=ERROR_COUNT_METRIC_NAME)
            )
            assert get_counter_value(p.run(), ERROR_COUNT_METRIC_NAME) == 1


@pytest.mark.slow
class TestTransformAndCount(BeamTest):
    def test_should_not_change_result(self):
        with TestPipeline() as p:
            result = (
                p |
                beam.Create([SOME_VALUE_1.lower()]) |
                TransformAndCount(
                    beam.Map(lambda x: x.upper()),
                    COUNT_METRIC_NAME_1
                )
            )
            assert_that(result, equal_to([SOME_VALUE_1.upper()]))

    def test_should_increase_count_per_item(self):
        with TestPipeline() as p:
            _ = (  # noqa: F841
                p |
                beam.Create([SOME_VALUE_1, SOME_VALUE_2]) |
                TransformAndCount(
                    beam.Map(lambda x: x),
                    COUNT_METRIC_NAME_1
                )
            )
            assert get_counter_value(p.run(), COUNT_METRIC_NAME_1) == 2

    def test_should_increase_count_per_item_using_function(self):
        with TestPipeline() as p:
            _ = (  # noqa: F841
                p |
                beam.Create([SOME_VALUE_1, SOME_VALUE_2]) |
                TransformAndCount(
                    beam.Map(lambda x: x),
                    COUNT_METRIC_NAME_1,
                    len
                )
            )
            assert get_counter_value(p.run(), COUNT_METRIC_NAME_1) == (
                len(SOME_VALUE_1) + len(SOME_VALUE_2)
            )


@pytest.mark.slow
class TestTransformAndLog(BeamTest):
    def test_should_not_change_result(self):
        with TestPipeline() as p:
            result = (
                p |
                beam.Create([SOME_VALUE_1.lower()]) |
                TransformAndLog(
                    beam.Map(lambda x: x.upper())
                )
            )
            assert_that(result, equal_to([SOME_VALUE_1.upper()]))


@pytest.mark.slow
class TestPreventFusion(BeamTest):
    def test_should_not_change_result_with_default_random_key(self):
        with TestPipeline() as p:
            result = (
                p |
                beam.Create([SOME_VALUE_1, SOME_VALUE_2]) |
                PreventFusion()
            )
            assert_that(result, equal_to([SOME_VALUE_1, SOME_VALUE_2]))

    def test_should_not_change_result_with_constant_key(self):
        with TestPipeline() as p:
            result = (
                p |
                beam.Create([SOME_VALUE_1, SOME_VALUE_2]) |
                PreventFusion(lambda _: 1)
            )
            assert_that(result, equal_to([SOME_VALUE_1, SOME_VALUE_2]))



from __future__ import absolute_import

import logging
from contextlib import contextmanager
from io import BytesIO
from abc import ABCMeta, abstractmethod

from unittest.mock import patch, Mock, MagicMock, MagicProxy
try:
    from mock import Mock as DeprecatedMock
    from mock import MagicMock as DeprecatedMagicMock
    from mock.mock import MagicProxy as DeprecatedMagicProxy
except ImportError:
    DeprecatedMock = None
    DeprecatedMagicMock = None
    DeprecatedMagicProxy = None

import pytest

from six import with_metaclass

import apache_beam as beam
from apache_beam.coders.coders import ToStringCoder, StrUtf8Coder
from apache_beam.testing.test_pipeline import TestPipeline as _TestPipeline
from apache_beam.io.filesystem import FileMetadata, MatchResult, CompressionTypes
from apache_beam.io.filesystems import FileSystems
from apache_beam.metrics.metric import MetricsFilter


class TestPipeline(_TestPipeline):
    __test__ = False

    def __init__(self, *args, **kwargs):
        super(TestPipeline, self).__init__(*args, **kwargs)
        self._pipeline_result = None

    def run(self):  # pylint: disable=arguments-differ
        # Make sure we're only running the pipeline once
        if not self._pipeline_result:
            self._pipeline_result = super(TestPipeline, self).run()
        return self._pipeline_result


_local = {}


def get_logger():
    return logging.getLogger(__name__)


class TestContext:
    def __init__(self):
        self.file_content_map = dict()
        self.object_map = dict()

    def set_file_content(self, name, content):
        get_logger().debug('set_file_content: %s (size: %d)', name, len(content))
        self.file_content_map[name] = content

    def get_file_content(self, name):
        return self.file_content_map.get(name)


def get_current_test_context():
    return _local['test_context']


# Apache Beam serialises everything, pretend Mocks being serialised


def unpickle_mock(state):
    # get_logger().debug('unpickle mock: state=%s', state)
    obj_id = state[0] if isinstance(state, tuple) else state
    obj = get_current_test_context().object_map[obj_id]
    return obj


unpickle_mock.__safe_for_unpickling__ = True


def mock_reduce(obj):
    obj_id = id(obj)
    # get_logger().debug('pickle mock, obj_id: %s', obj_id)
    get_current_test_context().object_map[obj_id] = obj
    return unpickle_mock, (obj_id,)


for c in [Mock, MagicMock, MagicProxy, DeprecatedMock, DeprecatedMagicMock, DeprecatedMagicProxy]:
    if c is None:
        continue
    c.__reduce__ = mock_reduce


@pytest.mark.filterwarnings('ignore::DeprecationWarning')
@pytest.mark.filterwarnings('ignore::UserWarning')
class BeamTest:
    @pytest.fixture(name='test_context', autouse=True)
    def init_test_context(self):
        get_logger().debug('setting up test context')
        test_context = TestContext()
        _local['test_context'] = test_context
        yield test_context
        get_logger().debug('clearing test context')
        del _local['test_context']


class MockWriteToText(beam.PTransform):
    class WriteDoFn(beam.DoFn):  # pylint: disable=abstract-method
        def __init__(self, file_path_prefix,
                     file_name_suffix='',
                     coder=ToStringCoder(),
                     header=None):
            super(MockWriteToText.WriteDoFn, self).__init__()
            self.filename = file_path_prefix + file_name_suffix
            self.file_obj = None
            self.coder = coder
            self.header = header

        def start_bundle(self):
            assert self.filename
            self.file_obj = BytesIO()
            if self.header:
                self.file_obj.write(self.coder.encode(self.header) + b'\n')

        def process(self, element):  # pylint: disable=arguments-differ
            assert self.file_obj
            self.file_obj.write(self.coder.encode(element) + b'\n')

        def finish_bundle(self):
            assert self.file_obj
            self.file_obj.flush()
            file_content = self.file_obj.getvalue()
            get_logger().debug('file content: %s: %s', self.filename, file_content)
            test_context = get_current_test_context()
            test_context.set_file_content(self.filename, file_content)
            self.file_obj.close()

    def __init__(self, *args, **kwargs):
        super(MockWriteToText, self).__init__()
        self._sink = MockWriteToText.WriteDoFn(*args, **kwargs)

    def expand(self, pcoll):  # pylint: disable=arguments-differ
        return pcoll | 'MockWriteToText' >> beam.ParDo(self._sink)


def MockReadFromText(
        file_pattern=None,
        coder=StrUtf8Coder(),
        skip_header_lines=0):

    file_content = get_current_test_context().get_file_content(file_pattern)
    if file_content is None:
        raise RuntimeError('no file content set for %s' % file_pattern)
    lines = file_content.replace('\r\n', '\n').split('\n')
    if skip_header_lines:
        lines = lines[skip_header_lines:]
    return 'MockReadFromText' >> beam.Create(
        [
            coder.decode(line)
            for line in lines
        ]
    )


class MockFileBasedSource(beam.io.filebasedsource.FileBasedSource):
    def open_file(self, file_name):
        file_content = get_current_test_context().get_file_content(file_name)
        if file_content is None:
            raise RuntimeError('no file content set for %s' % file_name)
        return BytesIO(file_content)


class AbstractFileSystem(with_metaclass(ABCMeta, object)):
    @abstractmethod
    def open(
            self, path, mime_type='application/octet-stream',
            compression_type=CompressionTypes.AUTO):
        pass

    @abstractmethod
    def create(
            self, path, mime_type='application/octet-stream',
            compression_type=CompressionTypes.AUTO):
        pass


class MockFileSystem(AbstractFileSystem):
    @classmethod
    def scheme(cls):
        return 'mock'

    def match(self, patterns, limits=None):
        test_context = get_current_test_context()
        file_content_map = test_context.file_content_map
        all_files = list(file_content_map.keys())
        if limits is None:
            limits = [None] * len(patterns)
        results = []
        for pattern, limit in zip(patterns, limits):
            files = all_files[:limit]
            metadata = [
                FileMetadata(f, len(file_content_map[f]))
                for f in files
            ]
            results.append(MatchResult(pattern, metadata))
        return results

    def open(
            self, path, mime_type='application/octet-stream',
            compression_type=CompressionTypes.AUTO):

        file_content = get_current_test_context().get_file_content(path)
        if file_content is None:
            raise RuntimeError('no file content set for %s' % path)
        return BytesIO(file_content)

    def create(
            self, path, mime_type='application/octet-stream',
            compression_type=CompressionTypes.AUTO):

        out = BytesIO()
        out.close = lambda: (
            get_current_test_context()
            .set_file_content(path, out.getvalue())
        )
        return out

    def rename(self, source_file_names, destination_file_names):
        test_context = get_current_test_context()
        file_content_map = test_context.file_content_map
        for source_file_name, destination_file_name in zip(
            source_file_names, destination_file_names
        ):
            get_logger().debug('renaming %s to %s', source_file_name, destination_file_name)
            if source_file_name not in file_content_map:
                raise IOError('mock file does not exist: %s' % source_file_name)
            if destination_file_name in file_content_map:
                raise IOError('mock file already exists: %s' % destination_file_name)
            file_content_map[destination_file_name] = file_content_map[source_file_name]
            del file_content_map[source_file_name]

    def mkdirs(self, path):
        get_logger().debug('mkdirs: %s (no-op)', path)


def mock_get_filesystem(*_):
    return MockFileSystem()


@contextmanager
def patch_beam_io():
    with patch.object(FileSystems, 'get_filesystem', classmethod(mock_get_filesystem)):
        yield


def get_counter_values(pipeline_result, names, wait_until_finish=True):
    if wait_until_finish:
        pipeline_result.wait_until_finish()
    counter_values = dict()
    for name in names:
        counter = pipeline_result.metrics().query(
            MetricsFilter().with_name(name)
        )['counters']
        assert len(counter) <= 1
        if len(counter) == 1:
            counter_values[name] = counter[0].committed
    return counter_values


def get_counter_value(pipeline_result, name, default_value=None, wait_until_finish=True):
    counter_values = get_counter_values(
        pipeline_result, [name], wait_until_finish=wait_until_finish
    )
    return counter_values.get(name, default_value)



import logging

import pytest


@pytest.fixture(scope='session', autouse=True)
def setup_logging():
    logging.basicConfig(level='INFO')
    for name in ['tests', 'sciencebeam_utils']:
        logging.getLogger(name).setLevel('DEBUG')



from mock import patch

import apache_beam as beam
from apache_beam.testing.util import assert_that, equal_to

from sciencebeam_utils.beam_utils.testing import (
    BeamTest,
    TestPipeline
)

import sciencebeam_utils.beam_utils.files as files_module
from sciencebeam_utils.beam_utils.files import (
    ReadFileList,
    DeferredReadFileList,
    FindFiles,
    DeferredFindFiles
)


FILE_1 = 'file1.pdf'
FILE_2 = 'file2.pdf'

FILE_LIST_PATH = 'file-list.lst'
COLUMN = 'url'
LIMIT = 10


class TestReadFileList(BeamTest):
    def test_should_use_load_file_list(self):
        with patch.object(files_module, 'load_file_list') as load_file_list:
            load_file_list.return_value = [FILE_1, FILE_2]
            with TestPipeline() as p:
                result = p | ReadFileList(FILE_LIST_PATH, column=COLUMN, limit=LIMIT)
                assert_that(result, equal_to([FILE_1, FILE_2]))
            load_file_list.assert_called_with(FILE_LIST_PATH, column=COLUMN, limit=LIMIT)


class TestDeferredReadFileList(BeamTest):
    def test_should_use_read_dict_csv(self):
        with patch.object(files_module, 'ReadDictCsv') as ReadDictCsv:
            ReadDictCsv.return_value = beam.Create([{COLUMN: FILE_1}, {COLUMN: FILE_2}])
            with TestPipeline() as p:
                result = p | DeferredReadFileList(FILE_LIST_PATH, column=COLUMN, limit=LIMIT)
                assert_that(result, equal_to([FILE_1, FILE_2]))
            ReadDictCsv.assert_called_with(FILE_LIST_PATH, limit=LIMIT)


class TestFindFiles(BeamTest):
    def test_should_use_find_matching_filenames(self):
        with patch.object(files_module, 'find_matching_filenames') as find_matching_filenames:
            find_matching_filenames.return_value = [FILE_1, FILE_2]
            with TestPipeline() as p:
                result = p | FindFiles(FILE_LIST_PATH, limit=LIMIT)
                assert_that(result, equal_to([FILE_1, FILE_2]))
            find_matching_filenames.assert_called_with(FILE_LIST_PATH)

    def test_should_apply_limit(self):
        with patch.object(files_module, 'find_matching_filenames') as find_matching_filenames:
            find_matching_filenames.return_value = [FILE_1, FILE_2]
            with TestPipeline() as p:
                result = p | FindFiles(FILE_LIST_PATH, limit=1)
                assert_that(result, equal_to([FILE_1]))


class TestDeferredFindFiles(BeamTest):
    def test_should_use_find_matching_filenames(self):
        with patch.object(files_module, 'find_matching_filenames') as find_matching_filenames:
            find_matching_filenames.return_value = [FILE_1, FILE_2]
            with TestPipeline() as p:
                result = p | DeferredFindFiles(FILE_LIST_PATH, limit=LIMIT)
                assert_that(result, equal_to([FILE_1, FILE_2]))
            find_matching_filenames.assert_called_with(FILE_LIST_PATH)

    def test_should_apply_limit(self):
        with patch.object(files_module, 'find_matching_filenames') as find_matching_filenames:
            find_matching_filenames.return_value = [FILE_1, FILE_2]
            with TestPipeline() as p:
                result = p | DeferredFindFiles(FILE_LIST_PATH, limit=1)
                assert_that(result, equal_to([FILE_1]))



from six import PY3


def python_2_unicode_compatible(cls):
    """
    Same as futures.utils.python_2_unicode_compatible but with support for __repr__
    """
    if not PY3:
        if cls.__repr__ is not object.__repr__:
            unicode_repr = cls.__repr__
            cls.__repr__ = lambda self: unicode_repr(self).encode('utf-8')
        if cls.__str__ is not object.__str__:
            cls.__unicode__ = cls.__str__
            cls.__str__ = lambda self: self.__unicode__().encode('utf-8')
    return cls



import argparse
import logging
import errno
from math import trunc
from random import shuffle
from datetime import datetime
from itertools import chain
from typing import List

from backports import csv  # pylint: disable=no-name-in-module

from six import text_type

from sciencebeam_utils.beam_utils.io import open_file

from sciencebeam_utils.utils.csv import (
    csv_delimiter_by_filename,
    write_csv_rows
)

from sciencebeam_utils.utils.file_path import (
    strip_ext,
    get_ext
)

from sciencebeam_utils.tools.tool_utils import (
    setup_logging,
    add_default_args,
    process_default_args
)


LOGGER = logging.getLogger(__name__)


Row = List[str]


def extract_proportions_from_args(args):
    digits = 3
    proportions = [
        (name, round(p, digits))
        for name, p in [
            ('train', args.train),
            ('test', args.test),
            ('validation', args.validation)
        ]
        if p and p > 0
    ]
    if sum(p for _, p in proportions) > 1.0:
        raise ValueError('proportions add up to more than 1.0')
    if not args.test:
        proportions.append(('test', 1.0 - sum(p for _, p in proportions)))
    elif not args.validation:
        proportions.append(('validation', round(1.0 - sum(p for _, p in proportions), digits)))
    proportions = [(name, p) for name, p in proportions if p > 0]
    return proportions


def get_chunk_size_list(size, percentages, fill=False):
    chunk_size_list = [int(trunc(p * size)) for p in percentages]
    if fill:
        chunk_size_list[-1] = size - sum(chunk_size_list[:-1])
    return chunk_size_list


def split_row_chunks(rows, chunk_size_list):
    chunk_offset_list = [0]
    for chunk_size in chunk_size_list[0:-1]:
        chunk_offset_list.append(chunk_offset_list[-1] + chunk_size)
    LOGGER.debug('chunk_offset_list: %s', chunk_offset_list)
    LOGGER.debug('chunk_size_list: %s', chunk_size_list)
    return [
        rows[chunk_offset:chunk_offset + chunk_size]
        for chunk_offset, chunk_size in zip(chunk_offset_list, chunk_size_list)
    ]


def _to_hashable(value):
    return tuple(value)


def _to_row_set(rows):
    return {_to_hashable(row) for row in rows}


def _split_rows_without_existing_split(rows, percentages, fill=False):
    chunk_size_list = get_chunk_size_list(len(rows), percentages, fill=fill)
    return split_row_chunks(rows, chunk_size_list)


def _substract_list(list1, list2):
    return [a - b for a, b in zip(list1, list2)]


def split_rows(
        rows: List[Row],
        percentages: List[float],
        fill: bool = False,
        existing_split: List[List[Row]] = None) -> List[List[Row]]:
    if not existing_split:
        return _split_rows_without_existing_split(rows, percentages, fill=fill)
    LOGGER.debug('existing_split: %s', existing_split)
    all_current_rows = _to_row_set(rows)
    all_existing_rows = _to_row_set(chain(*existing_split))
    not_existing_rows = all_existing_rows - all_current_rows
    if not_existing_rows:
        LOGGER.warning(
            'some rows (%d of %d) from the existing split do not exist'
            ' in the source list and will be removed, e.g.: %s',
            len(not_existing_rows), len(all_existing_rows), list(not_existing_rows)[:3]
        )
        existing_split = [
            [row for row in existing_rows if _to_hashable(row) in all_current_rows]
            for existing_rows in existing_split
        ]
    remaining_rows = [row for row in rows if _to_hashable(row) not in all_existing_rows]
    chunk_size_list = get_chunk_size_list(len(rows), percentages, fill=fill)
    existing_chunk_size_list = [len(existing_rows) for existing_rows in existing_split]
    remaining_chunk_size_list = _substract_list(chunk_size_list, existing_chunk_size_list)
    return [
        existing_rows + new_split
        for existing_rows, new_split in zip(
            existing_split, split_row_chunks(remaining_rows, remaining_chunk_size_list)
        )
    ]


def output_filenames_for_names(names, prefix, ext):
    return [
        prefix + ('' if prefix.endswith('/') else '-') + name + ext
        for name in names
    ]


def parse_args(argv=None):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--input', type=str, required=True,
        help='input csv/tsv file'
    )
    parser.add_argument(
        '--train', type=float, required=True,
        help='Train dataset proportion'
    )
    parser.add_argument(
        '--test', type=float, required=False,
        help='Test dataset proportion '
        '(if not specified it is assumed to be the remaining percentage)'
    )
    parser.add_argument(
        '--validation', type=float, required=False,
        help='Validation dataset proportion (requires test-proportion)'
    )
    parser.add_argument(
        '--random', action='store_true', default=False,
        help='randomise samples before doing the split'
    )
    parser.add_argument(
        '--fill', action='store_true', default=False,
        help='use up all of the remaining data rows for the last set'
    )
    parser.add_argument(
        '--no-extend-existing', action='store_true', default=False,
        help='do not extend and preserve the existing split (new entries will be addedby default)'
    )
    parser.add_argument(
        '--no-header', action='store_true', default=False,
        help='input file does not contain a header'
    )
    parser.add_argument(
        '--out', type=str, required=False,
        help='output csv/tsv file prefix or directory (if ending with slash)'
        ' will use input file name by default'
    )

    add_default_args(parser)

    return parser.parse_args(argv)


def process_args(args):
    if not args.out:
        args.out = strip_ext(args.input)


def read_csv_with_header(input_filename, delimiter, no_header):
    with open_file(input_filename, 'r') as f:
        reader = csv.reader(f, delimiter=text_type(delimiter))
        header_row = None if no_header else next(reader)
        data_rows = list(reader)
        return header_row, data_rows


def read_csv_data(input_filename, delimiter, no_header):
    _, data_rows = read_csv_with_header(input_filename, delimiter, no_header)
    return data_rows


def load_file_sets(filenames, delimiter, no_header):
    return [
        read_csv_data(filename, delimiter, no_header)
        for filename in filenames
    ]


def load_file_sets_or_none(filenames, delimiter, no_header):
    try:
        return load_file_sets(filenames, delimiter, no_header)
    except IOError as e:
        if e.errno == errno.ENOENT:
            return None
        raise e


def save_file_set(output_filename, delimiter, header_row, set_data_rows):
    mime_type = 'text/tsv' if delimiter == '\t' else 'text/csv'
    with open_file(output_filename, 'w', mime_type=mime_type) as f:
        writer = csv.writer(f, delimiter=text_type(delimiter))
        if header_row:
            write_csv_rows(writer, [header_row])
        write_csv_rows(writer, set_data_rows)


def save_file_sets(output_filenames, delimiter, header_row, data_rows_by_set):
    for output_filename, set_data_rows in zip(output_filenames, data_rows_by_set):
        LOGGER.info('set size: %d (%s)', len(set_data_rows), output_filename)
        save_file_set(output_filename, delimiter, header_row, set_data_rows)


def get_backup_file_suffix():
    return '.backup-%s' % datetime.utcnow().strftime(r'%Y%m%d-%H%M%S')


def run(args):
    LOGGER.debug('args: %s', args)
    process_args(args)
    ext = get_ext(args.input)
    proportions = extract_proportions_from_args(args)
    output_filenames = output_filenames_for_names(
        [name for name, _ in proportions],
        args.out,
        ext
    )

    LOGGER.info('proportions: %s', proportions)
    LOGGER.info('output_filenames: %s', output_filenames)

    delimiter = csv_delimiter_by_filename(args.input)

    header_row, data_rows = read_csv_with_header(args.input, delimiter, args.no_header)
    LOGGER.info('number of rows: %d', len(data_rows))

    if args.random:
        shuffle(data_rows)

    existing_file_sets = load_file_sets_or_none(output_filenames, delimiter, args.no_header)

    data_rows_by_set = split_rows(
        data_rows,
        [p for _, p in proportions],
        fill=args.fill,
        existing_split=existing_file_sets if not args.no_extend_existing else None
    )

    if existing_file_sets:
        backup_suffix = get_backup_file_suffix()
        save_file_sets(
            [s + backup_suffix for s in output_filenames],
            delimiter,
            header_row,
            existing_file_sets
        )

    save_file_sets(
        output_filenames,
        delimiter,
        header_row,
        data_rows_by_set
    )


def main(argv=None):
    args = parse_args(argv)

    process_default_args(args)

    run(args)


if __name__ == '__main__':
    setup_logging()

    main()



from __future__ import absolute_import

import os
import logging
from itertools import islice

from backports import csv  # pylint: disable=no-name-in-module

from six import text_type

from apache_beam.io.filesystems import FileSystems

from sciencebeam_utils.utils.csv import (
    csv_delimiter_by_filename
)

from sciencebeam_utils.beam_utils.io import open_file

from .file_path import (
    relative_path,
    join_if_relative_path
)


LOGGER = logging.getLogger(__name__)


def is_csv_or_tsv_file_list(file_list_path):
    return '.csv' in file_list_path or '.tsv' in file_list_path


def load_plain_file_list(file_list_path, limit=None):
    with open_file(file_list_path, 'r') as f:
        lines = (x.rstrip() for x in f)
        if limit:
            lines = islice(lines, 0, limit)
        return list(lines)


def load_csv_or_tsv_file_list(file_list_path, column, header=True, limit=None):
    delimiter = csv_delimiter_by_filename(file_list_path)
    with open_file(file_list_path, 'r') as f:
        reader = csv.reader(f, delimiter=text_type(delimiter))
        if not header:
            assert isinstance(column, int)
            column_index = column
        else:
            header_row = next(reader)
            if isinstance(column, int):
                column_index = column
            else:
                try:
                    column_index = header_row.index(column)
                except ValueError as exc:
                    raise ValueError(
                        'column %s not found, available columns: %s' %
                        (column, header_row)
                    ) from exc
        lines = (x[column_index] for x in reader)
        if limit:
            lines = islice(lines, 0, limit)
        return list(lines)


def to_absolute_file_list(base_path, file_list):
    return [join_if_relative_path(base_path, s) for s in file_list]


def to_relative_file_list(base_path, file_list):
    return [relative_path(base_path, s) for s in file_list]


def load_file_list(file_list_path, column, header=True, limit=None, to_absolute=True):
    if is_csv_or_tsv_file_list(file_list_path):
        file_list = load_csv_or_tsv_file_list(
            file_list_path, column=column, header=header, limit=limit
        )
    else:
        file_list = load_plain_file_list(file_list_path, limit=limit)
    if to_absolute:
        file_list = to_absolute_file_list(
            os.path.dirname(file_list_path), file_list
        )
    return file_list


def save_plain_file_list(file_list_path, file_list):
    with FileSystems.create(file_list_path) as f:
        f.write('\n'.join(file_list).encode('utf-8'))


def save_csv_or_tsv_file_list(file_list_path, file_list, column, header=True):
    if header:
        file_list = [column] + file_list
    save_plain_file_list(file_list_path, file_list)


def save_file_list(file_list_path, file_list, column, header=True):
    if is_csv_or_tsv_file_list(file_list_path):
        return save_csv_or_tsv_file_list(
            file_list_path, file_list, column=column, header=header
        )
    return save_plain_file_list(file_list_path, file_list)



from future.utils import python_2_unicode_compatible


@python_2_unicode_compatible
class LazyStr:
    def __init__(self, fn):
        self.fn = fn

    def __str__(self):
        return self.fn()


def parse_list(s, sep=','):
    s = s.strip()
    if not s:
        return []
    return [item.strip() for item in s.split(sep)]



from six import raise_from

import numpy as np


def assert_close(a, b, atol=1.e-8):
    try:
        assert np.allclose([a], [b], atol=atol)
    except AssertionError as e:
        raise_from(AssertionError('expected %s to be close to %s (atol=%s)' % (a, b, atol)), e)


def assert_all_close(a, b, atol=1.e-8):
    try:
        assert np.allclose(a, b, atol=atol)
    except AssertionError as e:
        raise_from(AssertionError('expected %s to be close to %s (atol=%s)' % (a, b, atol)), e)


def assert_all_not_close(a, b, atol=1.e-8):
    try:
        assert not np.allclose(a, b, atol=atol)
    except AssertionError as e:
        raise_from(AssertionError('expected %s not to be close to %s (atol=%s)' % (a, b, atol)), e)



from __future__ import absolute_import

import logging
from io import StringIO

from backports import csv  # pylint: disable=no-name-in-module

from six import text_type

import apache_beam as beam
from apache_beam.io.textio import WriteToText
from apache_beam.io.filesystem import CompressionTypes
from apache_beam.io.filebasedsource import FileBasedSource

from sciencebeam_utils.beam_utils.utils import (
    TransformAndLog
)

from sciencebeam_utils.utils.csv import (
    csv_delimiter_by_filename
)


def get_logger():
    return logging.getLogger(__name__)


def DictToList(fields):
    def wrapper(x):
        get_logger().debug('DictToList: %s -> %s', fields, x)
        return [x.get(field) for field in fields]
    return wrapper


def _to_text(value):
    try:
        return text_type(value, encoding='utf-8')
    except TypeError:
        return text_type(value)


def format_csv_rows(rows, delimiter=','):
    get_logger().debug('format_csv_rows, rows: %s', rows)
    out = StringIO()
    writer = csv.writer(out, delimiter=text_type(delimiter))
    writer.writerows([
        [_to_text(x) for x in row]
        for row in rows
    ])
    result = out.getvalue().rstrip('\r\n')
    get_logger().debug('format_csv_rows, result: %s', result)
    return result


class WriteDictCsv(beam.PTransform):
    def __init__(self, path, columns, file_name_suffix=None):
        super(WriteDictCsv, self).__init__()
        self.path = path
        self.columns = columns
        self.file_name_suffix = file_name_suffix
        self.delimiter = csv_delimiter_by_filename(path + file_name_suffix)

    def expand(self, pcoll):  # pylint: disable=arguments-differ
        return (
            pcoll |
            "ToList" >> beam.Map(DictToList(self.columns)) |
            "Format" >> TransformAndLog(
                beam.Map(lambda x: format_csv_rows([x], delimiter=self.delimiter)),
                log_prefix='formatted csv: ',
                log_level='debug'
            ) |
            "Utf8Encode" >> beam.Map(lambda x: x.encode('utf-8')) |
            "Write" >> WriteToText(
                self.path,
                file_name_suffix=self.file_name_suffix,
                header=format_csv_rows([self.columns], delimiter=self.delimiter).encode('utf-8')
            )
        )


def _strip_quotes(s):
    return s[1:-1] if len(s) >= 2 and s[0] == '"' and s[-1] == '"' else s

# copied and modified from https://github.com/pabloem/beam_utils
# (move back if still active)


class ReadLineIterator:
    def __init__(self, obj):
        self._obj = obj

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        line = self._obj.readline().decode('utf-8')
        if line is None or line == '':
            raise StopIteration
        return line


class CsvFileSource(FileBasedSource):
    """ A source for a GCS or local comma-separated-file
    Parses a text file assuming newline-delimited lines,
    and comma-delimited fields. Assumes UTF-8 encoding.
    """

    def __init__(  # pylint: disable=too-many-arguments
            self, file_pattern,
            compression_type=CompressionTypes.AUTO,
            delimiter=',', header=True, dictionary_output=True,
            validate=True, limit=None):
        """ Initialize a CsvFileSource.
        Args:
          delimiter: The delimiter character in the CSV file.
          header: Whether the input file has a header or not.
            Default: True
          dictionary_output: The kind of records that the CsvFileSource outputs.
            If True, then it will output dict()'s, if False it will output list()'s.
            Default: True
        Raises:
          ValueError: If the input arguments are not consistent.
        """
        super(CsvFileSource, self).__init__(
            file_pattern,
            compression_type=compression_type,
            validate=validate,
            splittable=False  # Can't just split anywhere
        )
        self.delimiter = delimiter
        self.header = header
        self.dictionary_output = dictionary_output
        self.limit = limit
        self._file = None

        if not self.header and dictionary_output:
            raise ValueError(
                'header is required for the CSV reader to provide dictionary output'
            )

    def read_records(self, file_name, range_tracker):  # noqa: E501 pylint: disable=arguments-differ,unused-argument
        # If a multi-file pattern was specified as a source then make sure the
        # start/end offsets use the default values for reading the entire file.
        headers = None
        self._file = self.open_file(file_name)

        reader = csv.reader(ReadLineIterator(self._file), delimiter=text_type(self.delimiter))

        line_no = 0
        for i, row in enumerate(reader):
            if self.header and i == 0:
                headers = row
                continue

            if self.limit and line_no >= self.limit:
                break

            line_no += 1
            if self.dictionary_output:
                yield dict(zip(headers, row))
            else:
                yield row


class ReadDictCsv(beam.PTransform):
    """
    Simplified CSV parser, which does not support:
    * multi-line values
    * delimiter within value
    """

    def __init__(self, filename, header=True, limit=None):
        super(ReadDictCsv, self).__init__()
        if not header:
            raise RuntimeError('header required')
        self.filename = filename
        self.columns = None
        self.delimiter = csv_delimiter_by_filename(filename)
        self.limit = limit
        self.row_num = 0

    def expand(self, pcoll):  # pylint: disable=arguments-differ
        return (
            pcoll |
            beam.io.Read(CsvFileSource(
                self.filename,
                delimiter=self.delimiter,
                limit=self.limit
            ))
        )



from six import text_type, PY2

from sciencebeam_utils.utils.compat import (
    python_2_unicode_compatible
)


ASCII_VALUE = 'abc'
UNICODE_VALUE = u'a\u1234b'
UNICODE_STR = UNICODE_VALUE.encode('utf-8') if PY2 else UNICODE_VALUE


@python_2_unicode_compatible
class ReprWrapper:
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return self.value


@python_2_unicode_compatible
class StrWrapper:
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value


@python_2_unicode_compatible
class ReprStrWrapper:
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return self.value

    def __str__(self):
        return self.value


class TestPython2UnicodeCompatible:
    def test_should_return_repr_ascii_value(self):
        assert repr(ReprWrapper(text_type(ASCII_VALUE))) == ASCII_VALUE

    def test_should_encode_repr_unicode_value_without_str(self):
        assert repr(ReprWrapper(UNICODE_VALUE)) == UNICODE_STR

    def test_should_encode_repr_unicode_value_with_str(self):
        assert repr(ReprStrWrapper(UNICODE_VALUE)) == UNICODE_STR

    def test_should_return_str_ascii_value(self):
        assert str(StrWrapper(text_type(ASCII_VALUE))) == ASCII_VALUE

    def test_should_encode_str_unicode_value_without_repr(self):
        assert str(StrWrapper(UNICODE_VALUE)) == UNICODE_STR

    def test_should_encode_str_unicode_value_with_repr(self):
        assert str(ReprStrWrapper(UNICODE_VALUE)) == UNICODE_STR

    def test_should_encode_str_unicode_value_with_repr_but_without_str(self):
        assert str(ReprWrapper(UNICODE_VALUE)) == UNICODE_STR



import argparse
import logging

from backports import csv  # pylint: disable=no-name-in-module

from six import text_type

from sciencebeam_utils.utils.csv import (
    csv_delimiter_by_filename,
    write_csv_rows
)

from sciencebeam_utils.beam_utils.io import (
    open_file,
    dirname,
    mkdirs_if_not_exists
)

from sciencebeam_utils.utils.file_path import (
    join_if_relative_path,
    relative_path
)

from sciencebeam_utils.utils.file_pairs import (
    find_file_pairs_grouped_by_parent_directory_or_name
)

from sciencebeam_utils.tools.tool_utils import (
    setup_logging,
    add_default_args,
    process_default_args
)


LOGGER = logging.getLogger(__name__)


def parse_args(argv=None):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--data-path', type=str, required=True,
        help='base data path'
    )
    parser.add_argument(
        '--source-pattern', type=str, required=True,
        help='source pattern'
    )
    parser.add_argument(
        '--xml-pattern', type=str, required=True,
        help='xml pattern'
    )
    parser.add_argument(
        '--out', type=str, required=True,
        help='output csv/tsv file'
    )

    parser.add_argument(
        '--use-relative-paths', action='store_true',
        help='create a file list with relative paths (relative to the data path)'
    )

    add_default_args(parser)

    return parser.parse_args(argv)


def save_file_pairs_to_csv(output_path, source_xml_pairs):
    mkdirs_if_not_exists(dirname(output_path))
    delimiter = csv_delimiter_by_filename(output_path)
    mime_type = 'text/tsv' if delimiter == '\t' else 'text/csv'
    with open_file(output_path, 'w', mime_type=mime_type) as f:
        writer = csv.writer(f, delimiter=text_type(delimiter))
        write_csv_rows(writer, [['source_url', 'xml_url']])
        write_csv_rows(writer, source_xml_pairs)
    LOGGER.info('written results to %s', output_path)


def to_relative_file_pairs(base_path, file_pairs):
    return (
        (relative_path(base_path, source_url), relative_path(base_path, xml_url))
        for source_url, xml_url in file_pairs
    )


def run(args):
    LOGGER.info('finding file pairs')
    source_xml_pairs = find_file_pairs_grouped_by_parent_directory_or_name([
        join_if_relative_path(args.data_path, args.source_pattern),
        join_if_relative_path(args.data_path, args.xml_pattern)
    ])

    if args.use_relative_paths:
        source_xml_pairs = to_relative_file_pairs(args.data_path, source_xml_pairs)

    source_xml_pairs = list(source_xml_pairs)

    save_file_pairs_to_csv(args.out, source_xml_pairs)


def main(argv=None):
    args = parse_args(argv)

    process_default_args(args)

    run(args)


if __name__ == '__main__':
    setup_logging()

    main()



import logging
from io import StringIO

from sciencebeam_utils.utils.progress_logger import logging_tqdm


class TestLoggingTqdm:
    def test_should_log_tqdm_output(self):
        logger = logging.Logger('test')
        out = StringIO()
        stream_handler = logging.StreamHandler(out)
        logger.addHandler(stream_handler)
        with logging_tqdm(total=2, logger=logger, mininterval=0) as pbar:
            pbar.update(1)
        last_log_line = out.getvalue().splitlines()[-1]
        assert '50%' in last_log_line
        assert '1/2' in last_log_line



import os
from tempfile import NamedTemporaryFile
from mock import patch
from backports.tempfile import TemporaryDirectory

import pytest

import sciencebeam_utils.utils.file_list as file_list_loader
from sciencebeam_utils.utils.file_list import (
    is_csv_or_tsv_file_list,
    load_plain_file_list,
    load_csv_or_tsv_file_list,
    to_absolute_file_list,
    to_relative_file_list,
    load_file_list,
    save_plain_file_list,
    save_csv_or_tsv_file_list,
    save_file_list
)


FILE_1 = 'file1.pdf'
FILE_2 = 'file2.pdf'
UNICODE_FILE_1 = u'file1\u1234.pdf'
FILE_LIST = [FILE_1, FILE_2]


@pytest.fixture(name='load_plain_file_list_mock')
def _load_plain_file_list():
    with patch.object(file_list_loader, 'load_plain_file_list') as mock:
        yield mock


@pytest.fixture(name='load_csv_or_tsv_file_list_mock')
def _load_csv_or_tsv_file_list():
    with patch.object(file_list_loader, 'load_csv_or_tsv_file_list') as mock:
        yield mock


@pytest.fixture(name='to_absolute_file_list_mock')
def _to_absolute_file_list():
    with patch.object(file_list_loader, 'to_absolute_file_list') as mock:
        yield mock


class TestIsCsvOrTsvFileList:
    def test_should_return_true_if_file_ext_is_csv(self):
        assert is_csv_or_tsv_file_list('files.csv')

    def test_should_return_true_if_file_ext_is_csv_gz(self):
        assert is_csv_or_tsv_file_list('files.csv.gz')

    def test_should_return_true_if_file_ext_is_tsv(self):
        assert is_csv_or_tsv_file_list('files.tsv')

    def test_should_return_true_if_file_ext_is_tsv_gz(self):
        assert is_csv_or_tsv_file_list('files.tsv.gz')

    def test_should_return_false_if_file_ext_is_lst(self):
        assert not is_csv_or_tsv_file_list('files.lst')

    def test_should_return_false_if_file_ext_is_lst_gz(self):
        assert not is_csv_or_tsv_file_list('files.lst.gz')


class TestLoadPlainFileList:
    def test_should_read_multiple_file_paths_from_file(self):
        with NamedTemporaryFile('w') as f:
            f.write('\n'.join([FILE_1, FILE_2]))
            f.flush()
            assert load_plain_file_list(f.name) == [FILE_1, FILE_2]

    def test_should_read_unicode_file(self):
        with NamedTemporaryFile('wb') as f:
            f.write('\n'.join([UNICODE_FILE_1]).encode('utf-8'))
            f.flush()
            assert load_plain_file_list(f.name) == [UNICODE_FILE_1]

    def test_should_apply_limit(self):
        with NamedTemporaryFile('w') as f:
            f.write('\n'.join([FILE_1, FILE_2]))
            f.flush()
            assert load_plain_file_list(f.name, limit=1) == [FILE_1]


class TestLoadCsvOrTsvFileList:
    def test_should_read_multiple_file_paths_from_file_with_header_using_column_name(self):
        with NamedTemporaryFile('w') as f:
            f.write('\n'.join(['url', FILE_1, FILE_2]))
            f.flush()
            assert load_csv_or_tsv_file_list(f.name, 'url') == [FILE_1, FILE_2]

    def test_should_read_multiple_file_paths_from_file_with_header_using_column_index(self):
        with NamedTemporaryFile('w') as f:
            f.write('\n'.join(['url', FILE_1, FILE_2]))
            f.flush()
            assert load_csv_or_tsv_file_list(f.name, 0) == [FILE_1, FILE_2]

    def test_should_read_multiple_file_paths_from_file_without_header(self):
        with NamedTemporaryFile('w') as f:
            f.write('\n'.join([FILE_1, FILE_2]))
            f.flush()
            assert load_csv_or_tsv_file_list(f.name, 0, header=False) == [FILE_1, FILE_2]

    def test_should_read_unicode_file(self):
        with NamedTemporaryFile('wb') as f:
            f.write('\n'.join(['url', UNICODE_FILE_1]).encode('utf-8'))
            f.flush()
            assert load_csv_or_tsv_file_list(f.name, 'url') == [UNICODE_FILE_1]

    def test_should_raise_exception_if_column_name_is_invalid(self):
        with pytest.raises(ValueError):
            with NamedTemporaryFile('w') as f:
                f.write('\n'.join(['url', FILE_1, FILE_2]))
                f.flush()
                assert load_csv_or_tsv_file_list(f.name, 'xyz') == [FILE_1, FILE_2]

    def test_should_raise_exception_if_column_index_is_invalid(self):
        with pytest.raises(IndexError):
            with NamedTemporaryFile('w') as f:
                f.write('\n'.join(['url', FILE_1, FILE_2]))
                f.flush()
                assert load_csv_or_tsv_file_list(f.name, 1) == [FILE_1, FILE_2]

    def test_should_apply_limit(self):
        with NamedTemporaryFile('w') as f:
            f.write('\n'.join(['url', FILE_1, FILE_2]))
            f.flush()
            assert load_csv_or_tsv_file_list(f.name, 'url', limit=1) == [FILE_1]


class TestToAbsoluteFileList:
    def test_should_make_path_absolute(self):
        assert to_absolute_file_list('/base/path', ['sub/file1']) == ['/base/path/sub/file1']

    def test_should_not_change_absolute_paths(self):
        assert to_absolute_file_list('/base/path', ['/other/file1']) == ['/other/file1']


class TestToRelativeFileList:
    def test_should_make_path_absolute(self):
        assert to_relative_file_list('/base/path', ['/base/path/sub/file1']) == ['sub/file1']

    def test_should_not_change_path_outside_base_path(self):
        assert to_relative_file_list('/base/path', ['/other/file1']) == ['/other/file1']


@pytest.mark.usefixtures(
    'load_plain_file_list_mock', 'load_csv_or_tsv_file_list_mock', 'to_absolute_file_list_mock'
)
class TestLoadFileList:
    def test_should_call_load_plain_file_list(self, load_plain_file_list_mock):
        result = load_file_list(
            'file-list.lst', column='url', header=True, limit=1, to_absolute=False
        )
        load_plain_file_list_mock.assert_called_with('file-list.lst', limit=1)
        assert result == load_plain_file_list_mock.return_value

    def test_should_call_load_csv_or_tsv_file_list(self, load_csv_or_tsv_file_list_mock):
        result = load_file_list(
            'file-list.csv', column='url', header=True, limit=1, to_absolute=False
        )
        load_csv_or_tsv_file_list_mock.assert_called_with(
            'file-list.csv', column='url', header=True, limit=1
        )
        assert result == load_csv_or_tsv_file_list_mock.return_value

    def test_should_make_file_list_absolute(
            self, load_plain_file_list_mock, to_absolute_file_list_mock):

        result = load_file_list('/base/path/file-list.lst', column='url', to_absolute=True)
        to_absolute_file_list_mock.assert_called_with(
            '/base/path', load_plain_file_list_mock.return_value
        )
        assert result == to_absolute_file_list_mock.return_value


class TestSavePlainFileList:
    def test_should_write_multiple_file_paths(self):
        with TemporaryDirectory() as path:
            file_list_path = os.path.join(path, 'out.lst')
            save_plain_file_list(file_list_path, [FILE_1, FILE_2])
            assert load_plain_file_list(file_list_path) == [FILE_1, FILE_2]

    def test_should_write_unicode_file(self):
        with TemporaryDirectory() as path:
            file_list_path = os.path.join(path, 'out.lst')
            save_plain_file_list(file_list_path, [UNICODE_FILE_1])
            assert load_plain_file_list(file_list_path) == [UNICODE_FILE_1]


class TestSaveCsvOrTsvFileList:
    def test_should_write_multiple_file_paths(self):
        with TemporaryDirectory() as path:
            file_list_path = os.path.join(path, 'out.csv')
            save_csv_or_tsv_file_list(file_list_path, [FILE_1, FILE_2], column='url')
            assert load_csv_or_tsv_file_list(file_list_path, column='url') == [FILE_1, FILE_2]

    def test_should_write_unicode_file(self):
        with TemporaryDirectory() as path:
            file_list_path = os.path.join(path, 'out.lst')
            save_csv_or_tsv_file_list(file_list_path, [UNICODE_FILE_1], column='url')
            assert load_csv_or_tsv_file_list(file_list_path, column='url') == [UNICODE_FILE_1]


class TestSaveFileList:
    def test_should_call_save_plain_file_list(self):
        with patch.object(file_list_loader, 'save_plain_file_list') as mock:
            save_file_list('file-list.lst', FILE_LIST, column='url', header=True)
            mock.assert_called_with('file-list.lst', FILE_LIST)

    def test_should_call_save_csv_or_tsv_file_list(self):
        with patch.object(file_list_loader, 'save_csv_or_tsv_file_list') as mock:
            save_file_list('file-list.csv', FILE_LIST, column='url', header=True)
            mock.assert_called_with('file-list.csv', FILE_LIST, column='url', header=True)



from itertools import islice

import apache_beam as beam

from sciencebeam_utils.beam_utils.csv import (
    ReadDictCsv
)

from sciencebeam_utils.beam_utils.io import (
    find_matching_filenames
)

from sciencebeam_utils.beam_utils.utils import (
    GroupTransforms
)

from sciencebeam_utils.utils.file_list import (
    load_file_list
)


def find_matching_filenames_with_limit(pattern, limit=None):
    return islice(
        find_matching_filenames(pattern),
        limit
    )


def ReadFileList(file_list_path, column, limit=None):
    file_list = load_file_list(file_list_path, column=column, limit=limit)
    return beam.Create(file_list)


def DeferredReadFileList(file_list_path, column, limit=None):
    return GroupTransforms(lambda p: (
        p |
        "ReadFileUrls" >> ReadDictCsv(file_list_path, limit=limit) |
        "TranslateFileUrls" >> beam.Map(lambda row: row[column])
    ))


def FindFiles(file_pattern, limit=None):
    file_list = list(find_matching_filenames_with_limit(file_pattern, limit=limit))
    return beam.Create(file_list)


def DeferredFindFiles(file_pattern, limit=None):
    return GroupTransforms(lambda p: (
        p |
        beam.Create([file_pattern]) |
        "FindFiles" >> beam.FlatMap(
            lambda pattern: find_matching_filenames_with_limit(pattern, limit)
        )
    ))



from __future__ import absolute_import

import logging
import sys
from contextlib import contextmanager

from tqdm import tqdm


class TqdmLoggingHandler(logging.StreamHandler):
    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.write(msg)
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:  # noqa pylint: disable=bare-except
            self.handleError(record)


def _is_console_logging_handler(handler):
    return isinstance(handler, logging.StreamHandler) and handler.stream in {sys.stdout, sys.stderr}


def _get_console_formatter(handlers):
    for handler in handlers:
        if _is_console_logging_handler(handler):
            return handler.formatter
    return None


@contextmanager
def redirect_logging_to_tqdm(logger=None):
    if logger is None:
        logger = logging.root
    tqdm_handler = TqdmLoggingHandler()
    original_handlers = logger.handlers
    tqdm_handler.setFormatter(_get_console_formatter(original_handlers))
    try:
        logger.handlers = [
            handler
            for handler in logger.handlers
            if not _is_console_logging_handler(handler)
        ] + [tqdm_handler]
        yield
    finally:
        logger.handlers = original_handlers


@contextmanager
def tqdm_with_logging_redirect(*args, **kwargs):
    logger = kwargs.pop('logger', None)
    with tqdm(*args, **kwargs) as pbar:
        with redirect_logging_to_tqdm(logger=logger):
            yield pbar



from lxml.builder import E

from sciencebeam_utils.utils.xml import (
    get_text_content,
    get_immediate_text,
    xml_from_string_with_recover
)


SOME_VALUE_1 = 'some value1'
SOME_VALUE_2 = 'some value2'


class TestGetTextContent:
    def test_should_return_simple_text(self):
        node = E.parent(SOME_VALUE_1)
        assert get_text_content(node) == SOME_VALUE_1

    def test_should_return_text_of_child_element(self):
        node = E.parent(E.child(SOME_VALUE_1))
        assert get_text_content(node) == SOME_VALUE_1

    def test_should_return_text_of_child_element_and_preceeding_text(self):
        node = E.parent(SOME_VALUE_1, E.child(SOME_VALUE_2))
        assert get_text_content(node) == SOME_VALUE_1 + SOME_VALUE_2

    def test_should_return_text_of_child_element_and_trailing_text(self):
        node = E.parent(E.child(SOME_VALUE_1), SOME_VALUE_2)
        assert get_text_content(node) == SOME_VALUE_1 + SOME_VALUE_2

    def test_should_return_text_of_parent_excluding_children_to_exclude(self):
        child = E.child(SOME_VALUE_1)
        node = E.parent(child, SOME_VALUE_2)
        assert get_text_content(node, exclude=[child]) == SOME_VALUE_2


class TestGetImmediateText:
    def test_should_return_simple_text(self):
        node = E.parent(SOME_VALUE_1)
        assert get_immediate_text(node) == [SOME_VALUE_1]

    def test_should_not_return_text_of_child_element(self):
        node = E.parent(E.child(SOME_VALUE_1))
        assert get_immediate_text(node) == []


class TestXmlFromStringWithRecover:
    def test_should_parse_clean_xml(self):
        root = xml_from_string_with_recover('<root><child1>%s</child1></root>' % SOME_VALUE_1)
        node = root.find('child1')
        assert node is not None
        assert node.text == SOME_VALUE_1

    def test_should_parse_xml_with_unencoded_ampersand(self):
        value = 'A & B'
        root = xml_from_string_with_recover('<root><child1>%s</child1></root>' % value)
        node = root.find('child1')
        assert node is not None
        assert node.text == 'A  B'

    def test_should_parse_xml_with_unencoded_unknown_entity(self):
        value = 'A &unknown; B'
        root = xml_from_string_with_recover('<root><child1>%s</child1></root>' % value)
        node = root.find('child1')
        assert node is not None
        assert node.text == 'A  B'



import os
from stat import S_IXUSR


ZIP_UNIX_SYSTEM = 3


def make_executable(path):
    os.chmod(path, os.stat(path).st_mode | S_IXUSR)


def extract_all_with_permission(zf, target_dir):
    for info in zf.infolist():
        extracted_path = zf.extract(info, target_dir)

        if info.create_system == ZIP_UNIX_SYSTEM:
            unix_attributes = info.external_attr >> 16
            if unix_attributes:
                os.chmod(extracted_path, unix_attributes)


def extract_all_with_executable_permission(zf, target_dir):
    for info in zf.infolist():
        extracted_path = zf.extract(info, target_dir)

        if info.create_system == ZIP_UNIX_SYSTEM and os.path.isfile(extracted_path):
            unix_attributes = info.external_attr >> 16
            if unix_attributes & S_IXUSR:
                make_executable(extracted_path)



from __future__ import absolute_import

import codecs
import logging
from contextlib import contextmanager
from io import BytesIO

from apache_beam.io.filesystems import FileSystems


DEFAULT_BUFFER_SIZE = 4096 * 1024


def get_logger():
    return logging.getLogger(__name__)


@contextmanager
def open_file(path, mode='r', encoding='utf-8', **kwargs):
    if mode in ('r', 'w'):
        info = codecs.lookup(encoding)
        open_binary = FileSystems.open if mode == 'r' else FileSystems.create
        with open_binary(path, **kwargs) as fp:
            # Python 3 CSV package expects a text file
            yield codecs.StreamReaderWriter(fp, info.streamreader, info.streamwriter)
    else:
        raise ValueError('invalid mode: %s' % mode)


def read_all_from_path(path, buffer_size=DEFAULT_BUFFER_SIZE):
    with FileSystems.open(path) as f:
        out = BytesIO()
        while True:
            buf = f.read(buffer_size)
            if not buf:
                break
            out.write(buf)
        return out.getvalue()


def dirname(path):
    return FileSystems.split(path)[0]


def basename(path):
    return FileSystems.split(path)[1]


def find_matching_filenames(pattern):
    return (x.path for x in FileSystems.match([pattern])[0].metadata_list)


def mkdirs_if_not_exists(path):
    if not FileSystems.exists(path):
        try:
            get_logger().info('attempting to create directory: %s', path)
            FileSystems.mkdirs(path)
        except IOError:
            if not FileSystems.exists(path):
                raise


def save_file_content(output_filename, data):
    mkdirs_if_not_exists(dirname(output_filename))
    # Note: FileSystems.create transparently handles compression based on the file extension
    with FileSystems.create(output_filename) as f:
        f.write(data)
    return output_filename



from mock import patch

import pytest

import sciencebeam_utils.tools.check_file_list as check_file_list_module
from sciencebeam_utils.tools.check_file_list import (
    map_file_list_to_file_exists,
    format_file_list,
    format_file_exists_results,
    check_files_and_report_result
)


FILE_1 = 'file1'
FILE_2 = 'file2'


class TestMapFileListToFileExists:
    def test_should_return_single_file_exists(self):
        m = check_file_list_module
        with patch.object(m, 'FileSystems') as FileSystems:
            assert map_file_list_to_file_exists(
                [FILE_1]
            ) == [FileSystems.exists.return_value]
            FileSystems.exists.assert_called_with(FILE_1)


class TestFormatFileList:
    def test_should_format_multiple_files(self):
        assert (
            format_file_list([FILE_1, FILE_2]) ==
            "['%s', '%s']" % (FILE_1, FILE_2)
        )


class TestFormatFileExistsResults:
    def test_should_format_no_files(self):
        assert (
            format_file_exists_results([], []) ==
            'empty file list'
        )

    def test_should_format_all_files_exist(self):
        assert (
            format_file_exists_results([True, True], [FILE_1, FILE_2]) ==
            'files exist: 2 (100%), files missing: 0 (0%)'
        )

    def test_should_format_files_partially_exist(self):
        assert (
            format_file_exists_results([True, False], [FILE_1, FILE_2]) ==
            'files exist: 1 (50%), files missing: 1 (50%) (example missing: {})'.format(
                format_file_list([FILE_2])
            )
        )


class TestCheckFileListAndReportResults:
    def test_should_pass_file_list_to_format(self):
        m = check_file_list_module
        with patch.object(m, 'map_file_list_to_file_exists') as map_file_list_to_file_exists_mock:
            with patch.object(m, 'format_file_exists_results') as format_file_exists_results_mock:
                map_file_list_to_file_exists_mock.return_value = [True, True]
                check_files_and_report_result([FILE_1, FILE_2])
                map_file_list_to_file_exists_mock.assert_called_with([FILE_1, FILE_2])
                format_file_exists_results_mock.assert_called_with([True, True], [FILE_1, FILE_2])

    def test_should_raise_error_if_none_of_the_files_were_found(self):
        m = check_file_list_module
        with patch.object(m, 'map_file_list_to_file_exists') as map_file_list_to_file_exists_mock:
            with pytest.raises(AssertionError):
                map_file_list_to_file_exists_mock.return_value = [False, False]
                check_files_and_report_result([FILE_1, FILE_2])



import argparse
import logging

from sciencebeam_utils.utils.file_list import (
    load_file_list,
    save_file_list,
    to_relative_file_list
)

from sciencebeam_utils.utils.file_path import (
    join_if_relative_path,
    get_or_validate_base_path,
    get_output_file
)

from sciencebeam_utils.tools.check_file_list import (
    check_files_and_report_result
)

from sciencebeam_utils.tools.tool_utils import (
    setup_logging,
    add_limit_args,
    add_default_args,
    process_default_args
)


LOGGER = logging.getLogger(__name__)


def parse_args(argv=None):
    parser = argparse.ArgumentParser(
        'Get output files based on source files and suffix.'
    )

    source = parser.add_argument_group('source')
    source.add_argument(
        '--source-file-list', type=str, required=True,
        help='path to source file list (tsv/csv/lst)'
    )
    source.add_argument(
        '--source-file-column', type=str, required=False,
        default='url',
        help='csv/tsv column (ignored for plain file list)'
    )
    source.add_argument(
        '--source-base-path', type=str, required=False,
        help='base data path for source file urls'
    )

    output = parser.add_argument_group('output')
    output.add_argument(
        '--output-file-list', type=str, required=True,
        help='path to output file list (tsv/csv/lst)'
    )
    output.add_argument(
        '--output-file-column', type=str, required=False,
        default='url',
        help='csv/tsv column (ignored for plain file list)'
    )
    output.add_argument(
        '--output-file-suffix', type=str, required=False,
        help='file suffix (will be added to source urls after removing ext)'
    )
    output.add_argument(
        '--output-base-path', type=str, required=False,
        help='base output path (by default source base path with"-results" suffix)'
    )
    output.add_argument(
        '--use-relative-paths', action='store_true',
        help='create a file list with relative paths (relative to the output data path)'
    )

    add_limit_args(parser)

    parser.add_argument(
        '--check', action='store_true', default=False,
        help='check whether the output files exist'
    )
    parser.add_argument(
        '--check-limit', type=int, required=False,
        help='limit the files to check'
    )

    add_default_args(parser)

    return parser.parse_args(argv)


def get_output_file_list(file_list, source_base_path, output_base_path, output_file_suffix):
    return [
        get_output_file(filename, source_base_path, output_base_path, output_file_suffix)
        for filename in file_list
    ]


def run(opt):
    source_file_list = load_file_list(
        join_if_relative_path(
            opt.source_base_path,
            opt.source_file_list
        ),
        column=opt.source_file_column,
        limit=opt.limit
    )
    source_base_path = get_or_validate_base_path(
        source_file_list, opt.source_base_path
    )

    target_file_list = get_output_file_list(
        source_file_list, source_base_path, opt.output_base_path, opt.output_file_suffix
    )

    if opt.check:
        check_file_list = (
            target_file_list[:opt.check_limit] if opt.check_limit
            else target_file_list
        )
        LOGGER.info(
            'checking %d (out of %d) files...',
            len(check_file_list), len(target_file_list)
        )
        check_files_and_report_result(check_file_list)

    if opt.use_relative_paths:
        target_file_list = to_relative_file_list(opt.output_base_path, target_file_list)

    LOGGER.info(
        'saving file list (with %d files) to: %s',
        len(target_file_list), opt.output_file_list
    )
    save_file_list(
        opt.output_file_list,
        target_file_list,
        column=opt.output_file_column
    )


def process_args(args):
    if not args.output_base_path:
        args.output_base_path = args.source_base_path + '-results'


def main(argv=None):
    args = parse_args(argv)

    process_default_args(args)
    process_args(args)

    run(args)


if __name__ == '__main__':
    setup_logging()

    main()




from __future__ import absolute_import

import os

from apache_beam.io.filesystems import FileSystems


def get_ext(filename):
    name, ext = os.path.splitext(filename)
    if ext == '.gz':
        ext = get_ext(name) + ext
    return ext


def strip_ext(filename):
    # strip of gz, assuming there will be another extension before .gz
    if filename.endswith('.gz'):
        filename = filename[:-3]
    return os.path.splitext(filename)[0]


def relative_path(base_path, path):
    if not base_path:
        return path
    if not base_path.endswith('/'):
        base_path += '/'
    return path[len(base_path):] if path.startswith(base_path) else path


def is_relative_path(path):
    return not path.startswith('/') and '://' not in path


def join_if_relative_path(base_path, path):
    return (
        FileSystems.join(base_path, path)
        if base_path and is_relative_path(path)
        else path
    )


def change_ext(path, old_ext, new_ext):
    if old_ext is None:
        old_ext = os.path.splitext(path)[1]
        if old_ext == '.gz':
            path = path[:-len(old_ext)]
            old_ext = os.path.splitext(path)[1]
    if old_ext and path.endswith(old_ext):
        return path[:-len(old_ext)] + new_ext
    return path + new_ext


def get_output_file(filename, source_base_path, output_base_path, output_file_suffix):
    return FileSystems.join(
        output_base_path,
        change_ext(
            relative_path(source_base_path, filename),
            None, output_file_suffix
        )
    )


def base_path_for_file_list(file_list):
    common_prefix = os.path.commonprefix(file_list)
    i = max(common_prefix.rfind('/'), common_prefix.rfind('\\'))
    if i >= 0:
        return common_prefix[:i]
    return ''


def get_or_validate_base_path(file_list, base_path):
    common_path = base_path_for_file_list(file_list)
    if base_path:
        if not common_path.startswith(base_path):
            raise AssertionError(
                "invalid base path '%s', common path is: '%s'" % (base_path, common_path)
            )
        return base_path
    return common_path



from sciencebeam_utils.utils.file_pairs import group_file_pairs_by_parent_directory_or_name


class TestGroupFilePairsByParentDirectoryOrName:
    def test_should_return_empty_list_with_empty_input_file_lists(self):
        assert list(group_file_pairs_by_parent_directory_or_name([
            [],
            []
        ])) == []

    def test_should_group_single_file(self):
        assert list(group_file_pairs_by_parent_directory_or_name([
            ['parent1/file.x'],
            ['parent1/file.y']
        ])) == [('parent1/file.x', 'parent1/file.y')]

    def test_should_group_single_file_in_directory_with_different_names(self):
        assert list(group_file_pairs_by_parent_directory_or_name([
            ['parent1/file1.x'],
            ['parent1/file2.y']
        ])) == [('parent1/file1.x', 'parent1/file2.y')]

    def test_should_ignore_files_in_different_directories(self):
        assert list(group_file_pairs_by_parent_directory_or_name([
            ['parent1/file.x'],
            ['parent2/file.y']
        ])) == []

    def test_should_group_multiple_files_in_separate_parent_directories(self):
        assert list(group_file_pairs_by_parent_directory_or_name([
            ['parent1/file.x', 'parent2/file.x'],
            ['parent1/file.y', 'parent2/file.y']
        ])) == [
            ('parent1/file.x', 'parent1/file.y'),
            ('parent2/file.x', 'parent2/file.y')
        ]

    def test_should_group_multiple_files_in_same_parent_directory_with_same_name(self):
        assert list(group_file_pairs_by_parent_directory_or_name([
            ['parent1/file1.x', 'parent1/file2.x'],
            ['parent1/file1.y', 'parent1/file2.y']
        ])) == [
            ('parent1/file1.x', 'parent1/file1.y'),
            ('parent1/file2.x', 'parent1/file2.y')
        ]

    def test_should_group_multiple_files_in_same_parent_directory_with_same_name_gzipped(self):
        assert list(group_file_pairs_by_parent_directory_or_name([
            ['parent1/file1.x.gz', 'parent1/file2.x.gz'],
            ['parent1/file1.y.gz', 'parent1/file2.y.gz']
        ])) == [
            ('parent1/file1.x.gz', 'parent1/file1.y.gz'),
            ('parent1/file2.x.gz', 'parent1/file2.y.gz')
        ]



import logging


def setup_logging():
    logging.basicConfig(level='INFO')
    logging.getLogger('oauth2client').setLevel('WARNING')


def add_limit_args(parser):
    parser.add_argument(
        '--limit', type=int, required=False,
        help='limit the files to process'
    )


def add_default_args(parser):
    parser.add_argument(
        '--debug', action='store_true', default=False,
        help='enable debug output'
    )


def process_default_args(args):
    if args.debug:
        logging.getLogger().setLevel('DEBUG')



from lxml import etree


def _get_text_content_and_exclude(node, exclude):
    result = ''
    if node.text is not None:
        result += node.text
    result += ''.join([
        (
            _get_text_content_and_exclude(c, exclude)
            if c not in exclude
            else ''
        ) +
        (c.tail if c.tail is not None else '')
        for c in node.iterchildren()
    ])
    return result


def get_text_content(node, exclude=None):
    '''
    Strip tags and return text content
    '''
    if not exclude:
        return ''.join(node.itertext())
    return _get_text_content_and_exclude(node, exclude)


def get_immediate_text(node):
    return node.xpath('text()')


def get_text_content_list(nodes, exclude=None):
    return [get_text_content(node, exclude=exclude) for node in nodes]


def xml_from_string_with_recover(s):
    parser = etree.XMLParser(recover=True)
    return etree.fromstring(s, parser=parser)


def set_or_remove_attrib(attrib, name, value):
    if value is None:
        if name in attrib:
            del attrib[name]
    else:
        attrib[name] = value



import logging
from random import getrandbits

import apache_beam as beam
from apache_beam.metrics.metric import Metrics


def get_logger():
    return logging.getLogger(__name__)


def Spy(f):
    def spy_wrapper(x):
        f(x)
        return x
    return spy_wrapper


def MapSpy(f):
    return beam.Map(Spy(f))


def _default_exception_log_fn(exception, value):
    get_logger().warning(
        'caught exception (ignoring item): %s, input: %.100s...',
        exception, value, exc_info=exception
    )


def MapOrLog(fn, log_fn=None, error_count=None):
    if log_fn is None:
        log_fn = _default_exception_log_fn
    error_counter = (
        Metrics.counter('MapOrLog', error_count)
        if error_count
        else None
    )

    def wrapper(x):
        try:
            yield fn(x)
        except Exception as e:  # pylint: disable=broad-except
            if error_counter:
                error_counter.inc()
            log_fn(e, x)
    return beam.FlatMap(wrapper)


LEVEL_MAP = {
    'info': logging.INFO,
    'debug': logging.DEBUG
}


def Count(name, counter_value_fn):
    counter = Metrics.counter('Count', name)

    def wrapper(x):
        counter.inc(counter_value_fn(x) if counter_value_fn else 1)
        return x
    return name >> beam.Map(wrapper)


class GroupTransforms(beam.PTransform):
    """
    Convenience method to allow a PTransform for grouping purpose
    to be defined using a lambda function.
    (Completely unrelated to GroupBy transforms)
    """

    def __init__(self, expand_fn):
        super(GroupTransforms, self).__init__()
        self.expand_fn = expand_fn

    def expand(self, pcoll):  # pylint: disable=W0221
        return self.expand_fn(pcoll)


def TransformAndCount(transform, counter_name, counter_value_fn=None):
    return GroupTransforms(lambda pcoll: (
        pcoll |
        transform |
        "Count" >> Count(counter_name, counter_value_fn)
    ))


def _identity(x):
    return x


def _get_default_output_log_fn(log_level, log_prefix, log_value_fn):
    if log_value_fn is None:
        log_value_fn = _identity
    log_level = LEVEL_MAP.get(log_level, log_level)

    def _log_fn(x):
        get_logger().log(
            log_level, '%s%.50s...', log_prefix, log_value_fn(x)
        )
    return _log_fn


def TransformAndLog(transform, log_fn=None, log_prefix='', log_value_fn=None, log_level='info'):
    if log_fn is None:
        log_fn = _get_default_output_log_fn(log_level, log_prefix, log_value_fn)

    return GroupTransforms(lambda pcoll: (
        pcoll |
        transform |
        "Log" >> MapSpy(log_fn)
    ))


def random_key():
    return getrandbits(32)


def _default_random_key_fn(_):
    return random_key()


def PreventFusion(key_fn=None, name="PreventFusion"):
    """
    Prevents fusion to allow better distribution across workers.

    See:
    https://cloud.google.com/dataflow/service/dataflow-service-desc#preventing-fusion

    TODO Replace by: https://github.com/apache/beam/pull/4040
    """
    if key_fn is None:
        key_fn = _default_random_key_fn
    return name >> GroupTransforms(lambda pcoll: (
        pcoll |
        "AddKey" >> beam.Map(lambda x: (key_fn(x), x)) |
        "GroupByKey" >> beam.GroupByKey() |
        "Ungroup" >> beam.FlatMap(lambda element: element[1])
    ))



import argparse
import subprocess
from mock import patch

import pytest

from six import text_type

from sciencebeam_utils.utils.collection import extend_dict

import sciencebeam_utils.beam_utils.main as main_module
from sciencebeam_utils.beam_utils.main import (
    get_cloud_project,
    process_cloud_args,
    add_cloud_args
)


PROJECT_1 = 'project1'


DEFAULT_ARGS = dict(
    project=None,
    job_name=None,
    job_name_suffix=None,
    num_workers=10,
    cloud=False
)

DEFAULT_OUTPUT_PATH = '/tmp/output'


@pytest.fixture(name='check_output_mock')
def _check_output_mock():
    with patch.object(subprocess, 'check_output') as mock:
        yield mock


@pytest.fixture(name='get_cloud_project_mock')
def _get_cloud_project_mock():
    with patch.object(main_module, 'get_cloud_project') as mock:
        yield mock


class TestGetCloudProject:
    def test_should_return_project(self, check_output_mock):
        check_output_mock.return_value = PROJECT_1
        assert get_cloud_project() == PROJECT_1

    def test_should_return_text_type_if_check_output_returns_bytes(self, check_output_mock):
        check_output_mock.return_value = PROJECT_1.encode('utf-8')
        assert isinstance(get_cloud_project(), text_type)

    def test_should_return_text_type_if_check_output_text_type(self, check_output_mock):
        check_output_mock.return_value = text_type(PROJECT_1)
        assert isinstance(get_cloud_project(), text_type)


class TestAddCloudArgs:
    def test_should_accept_num_workers_with_underscore(self):
        args = add_cloud_args(argparse.ArgumentParser()).parse_args(['--num_workers=123'])
        assert args.num_workers == 123

    def test_should_accept_num_workers_with_hyphen(self):
        args = add_cloud_args(argparse.ArgumentParser()).parse_args(['--num-workers=123'])
        assert args.num_workers == 123

    def test_should_accept_max_workers(self):
        args = add_cloud_args(argparse.ArgumentParser()).parse_args(['--max-workers=123'])
        assert args.max_num_workers == 123

    def test_should_accept_job_name_with_underscore(self):
        args = add_cloud_args(argparse.ArgumentParser()).parse_args(['--job_name=job1'])
        assert args.job_name == 'job1'

    def test_should_accept_job_name_with_hyphen(self):
        args = add_cloud_args(argparse.ArgumentParser()).parse_args(['--job-name=job1'])
        assert args.job_name == 'job1'


@pytest.mark.usefixtures('get_cloud_project_mock')
class TestProcessCloudArgs:
    def test_should_use_get_cloud_project_if_project_is_empty(
            self, get_cloud_project_mock):
        get_cloud_project_mock.return_value = PROJECT_1
        args = argparse.Namespace(**extend_dict(DEFAULT_ARGS, {
            'cloud': True,
            'project': None
        }))
        process_cloud_args(args, output_path=DEFAULT_OUTPUT_PATH)
        assert args.project == PROJECT_1  # pylint: disable=no-member

    def test_should_not_call_get_cloud_project_if_cloud_is_false(
            self, get_cloud_project_mock):
        args = argparse.Namespace(**extend_dict(DEFAULT_ARGS, {
            'cloud': False
        }))
        process_cloud_args(args, output_path=DEFAULT_OUTPUT_PATH)
        get_cloud_project_mock.assert_not_called()

    def test_should_not_call_get_cloud_project_if_project_was_specified(
            self, get_cloud_project_mock):
        args = argparse.Namespace(**extend_dict(DEFAULT_ARGS, {
            'cloud': True,
            'project': PROJECT_1
        }))
        process_cloud_args(args, output_path=DEFAULT_OUTPUT_PATH)
        get_cloud_project_mock.assert_not_called()



from __future__ import absolute_import

import os
import logging

from backports import csv  # pylint: disable=no-name-in-module


LOGGER = logging.getLogger(__name__)


TEMP_FILE_SUFFIX = '.part'


def csv_delimiter_by_filename(filename):
    if '.tsv' in filename:
        return '\t'
    return ','


def open_csv_output(filename):
    return open(filename, 'w')


def write_csv_rows(writer, iterable):
    writer.writerows(iterable)


def write_csv_row(writer, row):
    write_csv_rows(writer, [row])


def write_csv(filename, columns, iterable, delimiter=None):
    if delimiter is None:
        delimiter = csv_delimiter_by_filename(filename)
    is_stdout = filename in {'stdout', '/dev/stdout'}
    temp_filename = (
        filename + TEMP_FILE_SUFFIX
        if is_stdout
        else filename
    )
    if not is_stdout and os.path.isfile(filename):
        os.remove(filename)
    with open_csv_output(temp_filename) as csv_f:
        writer = csv.writer(csv_f, delimiter=delimiter)
        write_csv_rows(writer, [columns])
        write_csv_rows(writer, iterable)
    if not is_stdout:
        os.rename(temp_filename, filename)


def iter_dict_to_list(iterable, fields):
    return (
        [item.get(field) for field in fields]
        for item in iterable
    )


def write_dict_csv(filename, columns, iterable, delimiter=None):
    write_csv(filename, columns, iter_dict_to_list(iterable, columns), delimiter=delimiter)



from __future__ import absolute_import

from contextlib import contextmanager
from mock import patch

import pytest

import apache_beam as beam
from apache_beam.testing.util import assert_that, equal_to

from sciencebeam_utils.beam_utils.testing import (
    TestPipeline,
    BeamTest,
    MockWriteToText,
    patch_beam_io
)

from sciencebeam_utils.beam_utils.csv import (
    WriteDictCsv,
    ReadDictCsv,
    format_csv_rows
)


MODULE_UNDER_TEST = 'sciencebeam_utils.beam_utils.csv'

UNICODE_STR_1 = u'file1\u1234.pdf'


@contextmanager
def patch_module_under_test(**kwargs):
    with patch.multiple(
        MODULE_UNDER_TEST,
        **kwargs
    ) as mocks:
        yield mocks


def to_csv(rows, delimiter):
    return (format_csv_rows(rows, delimiter).replace('\r\n', '\n') + '\n').encode('utf-8')


class TestFormatCsvRows:
    def test_should_format_empty_rows(self):
        assert format_csv_rows([]) == ''

    def test_should_format_single_unicode_cell(self):
        assert format_csv_rows([[UNICODE_STR_1]]) == UNICODE_STR_1

    def test_should_format_single_byte_cell(self):
        assert format_csv_rows([[UNICODE_STR_1.encode('utf-8')]]) == UNICODE_STR_1

    def test_should_format_single_int_cell(self):
        assert format_csv_rows([[123]]) == '123'

    def test_should_format_single_string_row(self):
        assert format_csv_rows([['abc', 'def']]) == 'abc,def'

    def test_should_format_multiple_string_rows(self):
        assert (
            format_csv_rows([['abc', 'def'], ['123', '456']]).splitlines()
            == ['abc,def', '123,456']
        )


@pytest.mark.slow
class TestWriteDictCsv(BeamTest):
    def test_should_write_tsv_with_header(self, test_context):
        with patch_module_under_test(WriteToText=MockWriteToText):
            with TestPipeline() as p:
                _ = (  # noqa: F841
                    p |
                    beam.Create([{
                        'a': 'a1',
                        'b': 'b1'
                    }]) |
                    WriteDictCsv(
                        '.temp/dummy',
                        ['a', 'b'],
                        '.tsv'
                    )
                )
            assert test_context.get_file_content('.temp/dummy.tsv') == to_csv([
                ['a', 'b'],
                ['a1', 'b1']
            ], '\t')


@pytest.mark.slow
class TestReadDictCsv(BeamTest):
    def test_should_read_rows_as_dict(self, test_context):
        with patch_beam_io():
            test_context.set_file_content('.temp/dummy.tsv', to_csv([
                ['a', 'b'],
                ['a1', 'b1']
            ], '\t'))

            with TestPipeline() as p:
                result = (
                    p |
                    ReadDictCsv('.temp/dummy.tsv')
                )
                assert_that(result, equal_to([{
                    'a': 'a1',
                    'b': 'b1'
                }]))

    def test_should_read_multiple(self, test_context):
        with patch_beam_io():
            test_context.set_file_content('.temp/dummy.tsv', to_csv([
                ['a', 'b'],
                ['a1', 'b1'],
                ['a2', 'b2'],
                ['a3', 'b3']
            ], '\t'))

            with TestPipeline() as p:
                result = (
                    p |
                    ReadDictCsv('.temp/dummy.tsv')
                )
                assert_that(result, equal_to([{
                    'a': 'a1',
                    'b': 'b1'
                }, {
                    'a': 'a2',
                    'b': 'b2'
                }, {
                    'a': 'a3',
                    'b': 'b3'
                }]))

    def test_should_limit_number_of_rows(self, test_context):
        with patch_beam_io():
            test_context.set_file_content('.temp/dummy.tsv', to_csv([
                ['a', 'b'],
                ['a1', 'b1'],
                ['a2', 'b2'],
                ['a3', 'b3']
            ], '\t'))

            with TestPipeline() as p:
                result = (
                    p |
                    ReadDictCsv('.temp/dummy.tsv', limit=2)
                )
                assert_that(result, equal_to([{
                    'a': 'a1',
                    'b': 'b1'
                }, {
                    'a': 'a2',
                    'b': 'b2'
                }]))



from sciencebeam_utils.utils.exceptions import (
    is_serializable,
    get_serializable_exception
)


class SerializableException(RuntimeError):
    pass


class NotSerializableException(RuntimeError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._not_pickable = lambda: 1


class TestIsSerializable:
    def test_should_return_true_for_serializable_exception(self):
        assert is_serializable(SerializableException())

    def test_should_return_false_for_not_serializable_exception(self):
        assert not is_serializable(NotSerializableException())


class TestGetPickableException:
    def test_should_return_passed_in_serializable_exception(self):
        exception = SerializableException()
        serializable_exception = get_serializable_exception(exception)
        assert serializable_exception == exception
        assert is_serializable(serializable_exception)

    def test_should_return_serializable_exception_replacement_without_cause(self):
        exception = NotSerializableException("not serializable")
        serializable_exception = get_serializable_exception(exception)
        assert serializable_exception != exception
        assert str(serializable_exception) == str(exception)
        assert serializable_exception.__cause__ is None
        assert is_serializable(serializable_exception)

    def test_should_preserve_traceback(self):
        try:
            raise NotSerializableException("not serializable")
        except NotSerializableException as _exception:
            exception = _exception
        assert exception.__traceback__ is not None
        serializable_exception = get_serializable_exception(exception)
        assert serializable_exception != exception
        assert serializable_exception.__traceback__ == exception.__traceback__

    def test_should_return_serializable_exception_replacement_with_serializable_cause(self):
        base_exception = SerializableException("base")
        exception = NotSerializableException("not serializable")
        exception.__cause__ = base_exception
        serializable_exception = get_serializable_exception(exception)
        assert serializable_exception != exception
        assert str(serializable_exception) == str(exception)
        assert serializable_exception.__cause__ == base_exception
        assert is_serializable(serializable_exception)

    def test_should_return_serializable_exception_replacement_with_not_serializable_cause(self):
        base_exception = NotSerializableException("base")
        exception = NotSerializableException("not serializable")
        exception.__cause__ = base_exception
        serializable_exception = get_serializable_exception(exception)
        assert serializable_exception != exception
        assert str(serializable_exception) == str(exception)
        assert serializable_exception.__cause__ != base_exception
        assert str(serializable_exception.__cause__) == str(base_exception)
        assert is_serializable(serializable_exception)

    def test_should_return_serializable_exception_replacement_with_not_serializable_context(self):
        base_exception = NotSerializableException("base")
        exception = NotSerializableException("not serializable")
        exception.__context__ = base_exception
        serializable_exception = get_serializable_exception(exception)
        assert serializable_exception != exception
        assert str(serializable_exception) == str(exception)
        assert serializable_exception.__context__ != base_exception
        assert str(serializable_exception.__context__) == str(base_exception)
        assert is_serializable(serializable_exception)



from __future__ import absolute_import

from collections import namedtuple
from itertools import groupby
from typing import Iterable, List, T

from six import iteritems


def flatten(list_: Iterable[T]) -> List[T]:
    return [item for sublist in list_ for item in sublist]


def iter_flatten(list_: Iterable[T]) -> Iterable[T]:
    return (item for sublist in list_ for item in sublist)


def filter_truthy(list_: Iterable[T]) -> List[T]:
    return [item for item in list_ if item]


def strip_all(list_of_strings):
    return [(s or '').strip() for s in list_of_strings if s]


def remove_key_from_dict(d, key):
    return {k: v for k, v in iteritems(d) if k != key}


def remove_keys_from_dict(d, keys_to_remove):
    if not keys_to_remove:
        return d
    return {
        k: v
        for k, v in iteritems(d)
        if k not in keys_to_remove
    }


def extract_from_dict(d, key, default_value=None):
    return d.get(key, default_value), remove_key_from_dict(d, key)


def extend_dict(d, *other_dicts, **kwargs):
    """
    example:

    extend_dict(d1, d2)

    is equivalent to Python 3 syntax:
    {
      **d1,
      **d2
    }
    """
    d = d.copy()
    for other_dict in other_dicts:
        d.update(other_dict)
    d.update(kwargs)
    return d


def groupby_to_dict(iterable, key):
    return {
        k: list(v)
        for k, v in groupby(iterable, key=key)
    }


def sort_and_groupby_to_dict(iterable, key):
    return groupby_to_dict(sorted(iterable, key=key), key)


def to_namedtuple(*args, **kwargs):
    name = kwargs.pop('name', 'Tuple')
    d = extend_dict(*list(args) + [kwargs])
    return namedtuple(name, d.keys())(**d)



from __future__ import absolute_import

import logging
import sys
from io import StringIO

from sciencebeam_utils.utils.tqdm import (
    redirect_logging_to_tqdm,
    tqdm_with_logging_redirect,
    TqdmLoggingHandler
)


LOGGER = logging.getLogger(__name__)


class TestRedirectLoggingToTqdm:
    def test_should_add_and_remove_tqdm_handler(self):
        logger = logging.Logger('test')
        with redirect_logging_to_tqdm(logger=logger):
            assert len(logger.handlers) == 1
            assert isinstance(logger.handlers[0], TqdmLoggingHandler)
        assert not logger.handlers

    def test_should_remove_and_restore_console_handlers(self):
        logger = logging.Logger('test')
        stderr_console_handler = logging.StreamHandler(sys.stderr)
        stdout_console_handler = logging.StreamHandler(sys.stderr)
        logger.handlers = [stderr_console_handler, stdout_console_handler]
        with redirect_logging_to_tqdm(logger=logger):
            assert len(logger.handlers) == 1
            assert isinstance(logger.handlers[0], TqdmLoggingHandler)
        assert logger.handlers == [stderr_console_handler, stdout_console_handler]

    def test_should_inherit_console_logger_formatter(self):
        logger = logging.Logger('test')
        formatter = logging.Formatter('custom: %(message)s')
        console_handler = logging.StreamHandler(sys.stderr)
        console_handler.setFormatter(formatter)
        logger.handlers = [console_handler]
        with redirect_logging_to_tqdm(logger=logger):
            assert logger.handlers[0].formatter == formatter

    def test_should_not_remove_stream_handlers_not_fot_stdout_or_stderr(self):
        logger = logging.Logger('test')
        stream_handler = logging.StreamHandler(StringIO())
        logger.addHandler(stream_handler)
        with redirect_logging_to_tqdm(logger=logger):
            assert len(logger.handlers) == 2
            assert logger.handlers[0] == stream_handler
            assert isinstance(logger.handlers[1], TqdmLoggingHandler)
        assert logger.handlers == [stream_handler]


class TestTqdmWithLoggingRedirect:
    def test_should_add_and_remove_handler_from_root_logger_by_default(self):
        original_handlers = list(logging.root.handlers)
        with tqdm_with_logging_redirect(total=1) as pbar:
            assert isinstance(logging.root.handlers[-1], TqdmLoggingHandler)
            LOGGER.info('test')
            pbar.update(1)
        assert logging.root.handlers == original_handlers

    def test_should_add_and_remove_handler_from_custom_logger(self):
        logger = logging.Logger('test')
        with tqdm_with_logging_redirect(total=1, logger=logger) as pbar:
            assert len(logger.handlers) == 1
            assert isinstance(logger.handlers[0], TqdmLoggingHandler)
            LOGGER.info('test')
            pbar.update(1)
        assert not logger.handlers



from collections import namedtuple
from datetime import datetime

from mock import patch

import pytest

from sciencebeam_utils.tools import split_csv_dataset as split_csv_dataset_module
from sciencebeam_utils.tools.split_csv_dataset import (
    extract_proportions_from_args,
    split_rows,
    output_filenames_for_names,
    get_backup_file_suffix,
    parse_args,
    run
)


@pytest.fixture(name='get_backup_file_suffix_mock')
def _get_backup_file_suffix():
    with patch.object(split_csv_dataset_module, 'get_backup_file_suffix') as m:
        yield m


def create_args(**kwargs):
    return namedtuple('args', kwargs.keys())(**kwargs)


class TestExtractProportionsFromArgs:
    def test_should_create_train_test_split_with_only_train_specified(self):
        assert extract_proportions_from_args(
            create_args(train=0.6, test=None, validation=None)
        ) == [('train', 0.6), ('test', 0.4)]

    def test_should_create_train_test_validation_split_with_train_and_test_specified(self):
        assert extract_proportions_from_args(
            create_args(train=0.6, test=0.3, validation=None)
        ) == [('train', 0.6), ('test', 0.3), ('validation', 0.1)]

    def test_should_not_add_validation_if_remaining_percentage_is_zero(self):
        assert extract_proportions_from_args(
            create_args(train=0.6, test=0.4, validation=None)
        ) == [('train', 0.6), ('test', 0.4)]


def _chars_to_rows(chars):
    return [[c] for c in chars]


def _chars_to_rows_list(*chars_list):
    return [_chars_to_rows(chars) for chars in chars_list]


class TestSplitRows:
    def test_should_split_train_test(self):
        assert split_rows(
            _chars_to_rows('0123456789'), [0.6, 0.4]
        ) == _chars_to_rows_list('012345', '6789')

    def test_should_split_train_test_validation(self):
        assert split_rows(
            _chars_to_rows('0123456789'), [0.6, 0.3, 0.1]
        ) == _chars_to_rows_list('012345', '678', '9')

    def test_should_round_down(self):
        assert split_rows(
            _chars_to_rows('0123456789X'), [0.6, 0.4]
        ) == _chars_to_rows_list('012345', '6789')

    def test_should_fill_last_chunk_if_enabled(self):
        assert split_rows(
            _chars_to_rows('0123456789X'), [0.6, 0.4], fill=True
        ) == _chars_to_rows_list('012345', '6789X')

    def test_should_add_new_items_to_existing_split_train_test(self):
        existing_split = _chars_to_rows_list('012', '34')
        assert split_rows(
            _chars_to_rows('0123456789'), [0.6, 0.4], existing_split=existing_split
        ) == _chars_to_rows_list('012' + '567', '34' + '89')

    def test_should_add_new_items_to_existing_uneven_split_train_test(self):
        existing_split = _chars_to_rows_list('012', '3')
        assert split_rows(
            _chars_to_rows('0123456789'), [0.6, 0.4], existing_split=existing_split
        ) == _chars_to_rows_list('012' + '456', '3' + '789')

    def test_should_remove_items_not_present_in_total_list(self):
        existing_split = _chars_to_rows_list('1x', '2y')
        assert split_rows(
            _chars_to_rows('1234'), [0.5, 0.5], existing_split=existing_split
        ) == _chars_to_rows_list('13', '24')


class TestGetOutputFilenamesForNames:
    def test_should_add_name_and_ext_with_path_sep_if_out_ends_with_slash(self):
        assert output_filenames_for_names(
            ['train', 'test'], 'out/', '.tsv'
        ) == ['out/train.tsv', 'out/test.tsv']

    def test_should_add_name_and_ext_with_hyphen_if_out_does_not_end_with_slash(self):
        assert output_filenames_for_names(
            ['train', 'test'], 'out', '.tsv'
        ) == ['out-train.tsv', 'out-test.tsv']


class TestGetBackupFileSuffix:
    @patch.object(split_csv_dataset_module, 'datetime')
    def test_should_return_backup_suffix_with_datetime(self, datetime_mock):
        datetime_mock.utcnow.return_value = datetime(2001, 2, 3, 4, 5, 6)
        assert get_backup_file_suffix() == '.backup-20010203-040506'


class TestRun:
    def test_should_split_train_test(self, tmpdir):
        file_list = tmpdir.join('file-list.tsv')
        train_file_list = tmpdir.join('file-list-train.tsv')
        test_file_list = tmpdir.join('file-list-test.tsv')
        file_list.write('\n'.join(['header', 'row1', 'row2', 'row3', 'row4']))

        args = parse_args([
            '--input', str(file_list),
            '--train', str(0.5)
        ])
        run(args)

        assert (
            train_file_list.read().splitlines() ==
            ['header', 'row1', 'row2']
        )

        assert (
            test_file_list.read().splitlines() ==
            ['header', 'row3', 'row4']
        )

    def test_should_split_train_test_with_existing_split(self, tmpdir):
        file_list = tmpdir.join('file-list.tsv')
        train_file_list = tmpdir.join('file-list-train.tsv')
        test_file_list = tmpdir.join('file-list-test.tsv')
        file_list.write('\n'.join(['header', 'row1', 'row2', 'row3', 'row4']))
        train_file_list.write('\n'.join(['header', 'row1']))
        test_file_list.write('\n'.join(['header', 'row2']))

        args = parse_args([
            '--input', str(file_list),
            '--train', str(0.5)
        ])
        run(args)

        assert (
            train_file_list.read().splitlines() ==
            ['header', 'row1', 'row3']
        )

        assert (
            test_file_list.read().splitlines() ==
            ['header', 'row2', 'row4']
        )

    def test_should_backup_existing_split(self, tmpdir, get_backup_file_suffix_mock):
        get_backup_file_suffix_mock.return_value = '.backup'
        file_list = tmpdir.join('file-list.tsv')
        train_file_list = tmpdir.join('file-list-train.tsv')
        test_file_list = tmpdir.join('file-list-test.tsv')
        file_list.write('\n'.join(['header', 'row1', 'row2', 'row3', 'row4']))
        train_file_list.write('\n'.join(['header', 'row1']))
        test_file_list.write('\n'.join(['header', 'row2']))

        args = parse_args([
            '--input', str(file_list),
            '--train', str(0.5)
        ])
        run(args)

        assert (
            tmpdir.join('file-list-train.tsv.backup').read().splitlines() ==
            ['header', 'row1']
        )

        assert (
            tmpdir.join('file-list-test.tsv.backup').read().splitlines() ==
            ['header', 'row2']
        )



import os
import errno


def makedirs(path, exists_ok=False):
    try:
        # Python 3
        os.makedirs(path, exists_ok=exists_ok)  # pylint: disable=unexpected-keyword-arg
    except TypeError:
        # Python 2
        try:
            os.makedirs(path)
        except OSError as e:
            if e.errno == errno.EEXIST and os.path.isdir(path) and exists_ok:
                pass
            else:
                raise



from setuptools import find_packages, setup

import sciencebeam_utils


with open('requirements.txt', 'r') as f:
    REQUIRED_PACKAGES = f.readlines()


with open('README.md', 'r') as f:
    long_description = f.read()


packages = find_packages()


setup(
    name='sciencebeam_utils',
    version=sciencebeam_utils.__version__,
    author="eLife Sciences Publications, Ltd",
    url="https://github.com/elifesciences/sciencebeam-utils",
    install_requires=REQUIRED_PACKAGES,
    packages=packages,
    include_package_data=True,
    description='ScienceBeam Utils',
    long_description=long_description,
    long_description_content_type='text/markdown'
)



from sciencebeam_utils.utils.string import (
    parse_list
)


class TestParseList:
    def test_should_return_empty_list_for_empty_string(self):
        assert parse_list('') == []

    def test_should_return_single_item(self):
        assert parse_list('abc') == ['abc']

    def test_should_return_multiple_items(self):
        assert parse_list('abc,def') == ['abc', 'def']

    def test_should_ignore_space(self):
        assert parse_list(' abc , def ') == ['abc', 'def']



import os
from mock import patch, ANY

import pytest

from sciencebeam_utils.tools import get_output_files
from sciencebeam_utils.tools.get_output_files import (
    get_output_file_list,
    run,
    parse_args,
    main
)


SOME_ARGV = [
    '--source-file-list=source.csv',
    '--output-file-list=output.csv',
    '--limit=10'
]

BASE_SOURCE_PATH = '/source'

FILE_1 = BASE_SOURCE_PATH + '/file1'
FILE_2 = BASE_SOURCE_PATH + '/file2'


@pytest.fixture(name='load_file_list_mock')
def _load_file_list():
    with patch.object(get_output_files, 'load_file_list') as m:
        m.return_value = [FILE_1, FILE_2]
        yield m


@pytest.fixture(name='get_output_file_list_mock')
def _get_output_file_list():
    with patch.object(get_output_files, 'get_output_file_list') as m:
        yield m


@pytest.fixture(name='save_file_list_mock')
def _save_file_list():
    with patch.object(get_output_files, 'save_file_list') as m:
        yield m


@pytest.fixture(name='check_files_and_report_result_mock')
def _check_files_and_report_result():
    with patch.object(get_output_files, 'check_files_and_report_result') as m:
        yield m


@pytest.fixture(name='to_relative_file_list_mock')
def _to_relative_file_list():
    with patch.object(get_output_files, 'to_relative_file_list') as m:
        yield m


class TestGetOutputFileList:
    def test_should_return_output_file_with_path_and_change_ext(self):
        assert get_output_file_list(
            ['/source/path/file.pdf'],
            '/source',
            '/output',
            '.xml'
        ) == ['/output/path/file.xml']


@pytest.mark.usefixtures(
    "load_file_list_mock", "get_output_file_list_mock", "save_file_list_mock",
    "to_relative_file_list_mock"
)
class TestRun:
    def test_should_pass_around_parameters(
            self,
            load_file_list_mock,
            get_output_file_list_mock,
            save_file_list_mock):

        load_file_list_mock.return_value = [FILE_1, FILE_2]
        opt = parse_args(SOME_ARGV)
        run(opt)
        load_file_list_mock.assert_called_with(
            opt.source_file_list,
            column=opt.source_file_column,
            limit=opt.limit
        )
        get_output_file_list_mock.assert_called_with(
            load_file_list_mock.return_value,
            BASE_SOURCE_PATH,
            opt.output_base_path,
            opt.output_file_suffix
        )
        save_file_list_mock.assert_called_with(
            opt.output_file_list,
            get_output_file_list_mock.return_value,
            column=opt.source_file_column
        )

    def test_should_make_file_list_absolute_if_it_is_relative(
            self,
            load_file_list_mock):

        opt = parse_args(SOME_ARGV)
        opt.source_base_path = BASE_SOURCE_PATH
        opt.source_file_list = 'source.tsv'
        run(opt)
        load_file_list_mock.assert_called_with(
            os.path.join(opt.source_base_path, opt.source_file_list),
            column=opt.source_file_column,
            limit=opt.limit
        )

    def test_should_raise_error_if_source_path_is_invalid(self):
        opt = parse_args(SOME_ARGV)
        opt.source_base_path = '/other/path'
        with pytest.raises(AssertionError):
            run(opt)

    def test_should_use_passed_in_source_path_if_valid(
            self,
            get_output_file_list_mock,
            load_file_list_mock):

        opt = parse_args(SOME_ARGV)
        opt.source_base_path = '/base'
        load_file_list_mock.return_value = ['/base/source/file1', '/base/source/file2']
        run(opt)
        get_output_file_list_mock.assert_called_with(
            ANY,
            opt.source_base_path,
            ANY,
            ANY
        )

    def test_should_check_file_list_if_enabled(
            self,
            get_output_file_list_mock,
            check_files_and_report_result_mock):

        opt = parse_args(SOME_ARGV)
        opt.check = True
        run(opt)
        check_files_and_report_result_mock.assert_called_with(
            get_output_file_list_mock.return_value
        )

    def test_should_limit_files_to_check(
            self,
            load_file_list_mock,
            get_output_file_list_mock,
            check_files_and_report_result_mock):

        opt = parse_args(SOME_ARGV)
        opt.check = True
        opt.check_limit = 1
        load_file_list_mock.return_value = [FILE_1, FILE_2]
        run(opt)
        check_files_and_report_result_mock.assert_called_with(
            get_output_file_list_mock.return_value[:opt.check_limit]
        )

    def test_should_save_relative_paths_if_enabled(
            self,
            get_output_file_list_mock,
            to_relative_file_list_mock,
            save_file_list_mock):

        opt = parse_args(SOME_ARGV)
        opt.use_relative_paths = True
        run(opt)
        to_relative_file_list_mock.assert_called_with(
            opt.output_base_path,
            get_output_file_list_mock.return_value,
        )
        save_file_list_mock.assert_called_with(
            opt.output_file_list,
            to_relative_file_list_mock.return_value,
            column=opt.source_file_column
        )


class TestMain:
    def test_should_parse_args_and_call_run(self):
        m = get_output_files
        with patch.object(m, 'parse_args') as parse_args_mock:
            with patch.object(m, 'run') as run_mock:
                main(SOME_ARGV)
                parse_args_mock.assert_called_with(SOME_ARGV)
                run_mock.assert_called_with(parse_args_mock.return_value)
